/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <sys/statvfs.h>

#include "dico.h"
#include "common.h"
#include "archive.h"
#include "queue.h"
#include "comp_gzip.h"
#include "comp_bzip2.h"

int archive_init(struct s_archive *ai)
{
	if (!ai)
	{	errprintf("ai is NULL\n");
		return -1;
	}
	
	memset(ai, 0, sizeof(struct s_archive));
	ai->archfd=-1;
	ai->archid=0;
	ai->locked=false;
	ai->curvol=0;
	return 0;
}

int archive_destroy(struct s_archive *ai)
{
	if (!ai)
	{	errprintf("ai is NULL\n");
		return -1;
	}
	
	return 0;
}

int archive_generate_id(struct s_archive *ai)
{
	if (!ai)
	{	errprintf("ai is NULL\n");
		return -1;
	}
	
	// generate an archive id
	ai->archid=generate_random_u32_id();
	
	return 0;
}

int archive_create(struct s_archive *ai)
{
	if (!ai)
	{	errprintf("ai is NULL\n");
		return -1;
	}
	
	// check the archive does not exist
	if ((g_options.overwrite==0) && (regfile_exists(ai->volpath)==true))
	{	errprintf("%s already exists, please remove it first.\n", ai->basepath);
		return -1;
	}
	
	ai->archfd=open64(ai->volpath, O_RDWR|O_CREAT|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (ai->archfd<0)
	{       sysprintf ("cannot create archive %s\n", ai->volpath);
		return -1;
	}
	
	if (lockf(ai->archfd, F_LOCK, 0)!=0)
	{	sysprintf("Cannot lock archive file: %s\n", ai->volpath);
		close(ai->archfd);
		return -1;
	}
	ai->locked=true;
	
	return 0;
}

int archive_open(struct s_archive *ai)
{	
	struct stat64 st;
	
	if (!ai)
	{	errprintf("ai is NULL\n");
		return -1;
	}
	
	ai->archfd=open64(ai->volpath, O_RDONLY|O_LARGEFILE);
	if (ai->archfd<0)
	{       sysprintf ("cannot open archive %s\n", ai->volpath);
		return -1;
	}
	
	if (fstat64(ai->archfd, &st)!=0)
	{	sysprintf("fstat64(%d) failed\n", ai->volpath);
		return -1;
	}
	
	if (!S_ISREG(st.st_mode))
	{	errprintf("%s is not a regular file, cannot continue\n", ai->volpath);
		close(ai->archfd);
		return -1;
	}
	
	return 0;
}

int archive_close(struct s_archive *ai)
{
	if (!ai)
	{	errprintf("ai is NULL\n");
		return -1;
	}
	
	if (ai->archfd<0)
	{	errprintf("invalid file descriptor: %d\n", (int)ai->archfd);
		return -1;
	}
	
	if (ai->locked)
		lockf(ai->archfd, F_ULOCK, 0);
	fsync(ai->archfd); // just in case the user reboots after it exits
	close(ai->archfd);
	ai->archfd=-1;

	return 0;
}

int archive_remove(struct s_archive *ai)
{
	if (!ai)
	{	errprintf("ai is NULL\n");
		return -1;
	}
	
	if (ai->archfd!=-1)
		archive_close(ai);
	unlink(ai->basepath);
	msgprintf(MSG_FORCE, "removing %s\n", ai->basepath);
	
	return 0;
}

s64 archive_get_currentpos(struct s_archive *ai)
{
	return (s64)lseek64(ai->archfd, 0, SEEK_CUR);
}

int archive_read_data(struct s_archive *ai, void *data, u64 size)
{
	long lres;
	
	if (!ai)
	{	errprintf("ai is NULL\n");
		return -1;
	}
	
	if ((lres=read(ai->archfd, (char*)data, (long)size))!=(long)size)
	{	sysprintf("read failed: read(size=%ld)=%ld\n", (long)size, lres);
		return -1;
	}
	
	return 0;
}

int archive_read_dico(struct s_archive *ai, struct s_dico *d)
{
	u16 size;
	u32 headerlen;
	u32 origsum;
	u32 newsum;
	u8 *buffer;
	u8 *bufpos;
	u16 temp16;
	u32 temp32;
	u64 temp64;
	u8 section;
	u16 count;
	u64 key;
	int i;
	
	if (!ai || !d)
	{	errprintf("null parameters\n");
		return ERR_FATAL;
	}
	
	// header-len, header-data, header-checksum
	if (archive_read_data(ai, &temp32, sizeof(temp32))!=0)
	{	errprintf("imgdisk_read_data() failed\n");
		return ERR_FATAL;
	}
	headerlen=le32_to_cpu(temp32);
	if (headerlen > FSA_MAX_HEADSIZE)
	{	errprintf("invalid headerlen: %ld\n", (long)headerlen);
		return ERR_FATAL;
	}
	
	bufpos=buffer=malloc(headerlen);
	if (!buffer)
	{	errprintf("cannot allocate memory for header\n");
		return ERR_FATAL;
	}
	
	if (archive_read_data(ai, buffer, headerlen)!=0)
	{	errprintf("cannot read header data\n");
		free(buffer);
		return ERR_FATAL;
	}
	
	if (archive_read_data(ai, &temp32, sizeof(temp32))!=0)
	{	errprintf("cannot read header checksum\n");
		free(buffer);
		return ERR_FATAL;
	}
	origsum=le32_to_cpu(temp32);
	
	// check header-data integrity using checksum	
	newsum=fletcher32(buffer, headerlen);
	
	if (newsum!=origsum)
	{	errprintf("bad checksum for header\n");
		free(buffer);
		return ERR_MINOR; // header corrupt --> skip file
	}
	
	// read count from buffer
	memcpy(&temp16, bufpos, sizeof(temp16));
	bufpos+=sizeof(temp16);
	count=le16_to_cpu(temp16);
	
	// read items
	for (i=0; i < count; i++)
	{
		// a. read section from buffer
		memcpy(&section, bufpos, sizeof(section));
		bufpos+=sizeof(section);
		
		// b. read key from buffer
		memcpy(&temp64, bufpos, sizeof(temp64));
		bufpos+=sizeof(temp64);
		key=(u64)le64_to_cpu(temp64);
		
		// c. read sizeof(data)
		memcpy(&temp16, bufpos, sizeof(temp16));
		bufpos+=sizeof(temp16);
		size=le16_to_cpu(temp16);
		if (size>DICO_MAX_DATASIZE) // invalid length
			return ERR_MINOR;
		
		// d. add item to dico
		if (dico_add_data(d, section, key, bufpos, size)!=0)
			return ERR_FATAL;
		bufpos+=size;
	}
	
	free(buffer);
	return ERR_SUCCESS;
}

int archive_read_header(struct s_archive *ai, char *magic, struct s_dico **d, bool allowseek)
{
	s64 curpos;
	u32 temp32;
	u32 archid;
	int res;
	
	// init
	memset(magic, 0, FSA_SIZEOF_MAGIC);
	*d=NULL;
	
	if (!ai)
	{	errprintf("ai is NULL\n");
		return ERR_FATAL;
	}
	
	if ((*d=dico_alloc())==NULL)
	{	errprintf("dico_alloc() failed\n");
		return ERR_FATAL;
	}
	
	// search for next read header marker and magic (it may be further if corruption in archive)
	if ((curpos=lseek64(ai->archfd, 0, SEEK_CUR))<0)
	{	sysprintf("lseek64() failed to get the current position in archive\n");
		return ERR_FATAL;
	}
	
	if ((res=archive_read_data(ai, magic, FSA_SIZEOF_MAGIC))!=ERR_SUCCESS)
	{	msgprintf(MSG_STACK, "cannot read header magic: res=%d\n", res);
		return ERR_FATAL;
	}
	
	// we don't want to search for the magic if it's a volume header
	if (is_magic_valid(magic)!=true && allowseek!=true)
	{	errprintf("cannot read header magic: this is not a valid fsarchiver file, or it has been created with a different version.\n");
		return ERR_FATAL;
	}
	
	while (is_magic_valid(magic)!=true)
	{
		if (lseek64(ai->archfd, curpos++, SEEK_SET)<0)
		{	sysprintf("lseek64(pos=%lld, SEEK_SET) failed\n", (long long)curpos);
			return ERR_FATAL;
		}
		if ((res=archive_read_data(ai, magic, FSA_SIZEOF_MAGIC))!=ERR_SUCCESS)
		{	msgprintf(MSG_STACK, "cannot read header magic: res=%d\n", res);
			return ERR_FATAL;
		}
	}
	
	// read the archive id
	if ((res=archive_read_data(ai, &temp32, sizeof(temp32)))!=ERR_SUCCESS)
	{	msgprintf(MSG_STACK, "cannot read archive-id in header: res=%d\n", res);
		return ERR_FATAL;
	}
	archid=le32_to_cpu(temp32);
	if (ai->archid) // only check archive-id if it's known (when main header has been read)
	{
		if (archid!=ai->archid)
		{	errprintf("archive-id in header does not match: archid=[%.8x], expected=[%.8x]\n", archid, ai->archid);
			return ERR_MINOR;
		}
	}
	
	// read the dico of the header
	if ((res=archive_read_dico(ai, *d))!=ERR_SUCCESS)
	{	msgprintf(MSG_STACK, "imgdisk_read_dico() failed\n");
		return res;
	}
	
	return ERR_SUCCESS;
}

int archive_write_data(struct s_archive *ai, void *data, u64 size)
{
	struct statvfs64 statvfsbuf;
	char textbuf[128];
	long lres;
	
	if (!ai || !data || !size)
	{	errprintf("a parameter is null\n");
		return -1;
	}
	
	if ((lres=write(ai->archfd, (char*)data, (long)size))!=(long)size)
	{
		errprintf("write(size=%ld) returned %ld\n", (long)size, (long)lres);
		if ((lres>0) && (lres < (long)size)) // probably "no space left"
		{
			if (fstatvfs64(ai->archfd, &statvfsbuf)!=0)
			{	sysprintf("fstatvfs(fd=%d) failed\n", ai->archfd);
				return -1;
			}
			
			u64 freebytes = statvfsbuf.f_bfree * statvfsbuf.f_bsize;
			errprintf("Can't write archive file: space on device is %s\n", 
				format_size(freebytes, textbuf, sizeof(textbuf), 'h'));
			return -1;
		}
		else // another error
		{
			sysprintf("write(size=%ld) failed\n", (long)size);
			return -1;
		}
	}
	
	return 0;
}

// calculates the volpath using basepath and curvol
int archive_volpath(struct s_archive *ai)
{
	char temp[PATH_MAX];
	int pathlen;
	
	if (!ai)
	{	errprintf("ai is null\n");
		return -1;
	}
	
	if ((pathlen=strlen(ai->basepath))<4) // all archives terminates with ".fsa"
	{	errprintf("archive has an invalid basepath: [%s]\n", ai->basepath);
		return -1;
	}
	
	memset(ai->volpath, 0, PATH_MAX);
	memcpy(ai->volpath, ai->basepath, pathlen-2);
	if (ai->curvol==0) // first volume
	{
		snprintf(ai->volpath, PATH_MAX, "%s", ai->basepath);
	}
	else // not the first volume
	{
		if (ai->curvol<100) // 1..99
			snprintf(temp, sizeof(temp), "%.2ld", (long)ai->curvol);
		else // >=100
			snprintf(temp, sizeof(temp), "%ld", (long)ai->curvol);
		bufcat(ai->volpath, temp, PATH_MAX);
	}
	
	return 0;
}

int archive_incvolume(struct s_archive *ai, bool waitkeypress)
{
	if (!ai)
	{	errprintf("ai is null\n");
		return -1;
	}
	
	ai->curvol++;
	
	return archive_volpath(ai);
}
