/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <stdarg.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <linux/limits.h>
#include <dirent.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "syncthread.h"
#include "common.h"

int stream_readline(FILE *f, char *buf, int buflen)
{
	int i;
	char c;

	memset(buf, 0, buflen);
	for (i=0; (!feof(f)) && ((c=fgetc(f))!='\n') && (!feof(f)) && (i < buflen); i++)
		buf[i]=c;
	return i;
}

void concatenate_paths(char *buffer, int maxbufsize, char *p1, char *p2)
{
	int i;

	memset(buffer, 0, maxbufsize);
	for (i=0; (p1[i]) && (p1[i]!='/' || p1[i+1]); i++)
		*(buffer++)=p1[i];
	*(buffer++)='/';
	while (*p2=='/')
		p2++;
	bufcat(buffer, p2, maxbufsize);
}

char *format_size(u64 size, char *text, int max, char units)
{	
	double dSize;
	
	u64 llKiloB = 1024LL;
	u64 llMegaB = 1024LL * llKiloB;
	u64 llGigaB = 1024LL * llMegaB;
	u64 llTeraB = 1024LL * llGigaB;
	
	if ((units=='b') || ((units=='h') && size < llKiloB)) // In Bytes
	{	snprintf(text, max, "%lld bytes", (long long)size);
	}
	else if ((units=='k') || ((units=='h') && size < llMegaB)) // In KiloBytes
	{	dSize = ((double)size) / ((double)llKiloB);
		snprintf(text, max, "%.2f KB", (float) dSize);
	}
	else if ((units=='m') || ((units=='h') && size < llGigaB)) // In MegaBytes
	{	dSize = ((double)size) / ((double)llMegaB);
		snprintf(text, max, "%.2f MB", (float) dSize);
	}
	else if ((units=='g') || ((units=='h') && size < llTeraB)) // In Gigabytes
	{	dSize = ((double)size) / ((double)llGigaB);
		snprintf(text, max, "%.2f GB", (float) dSize);
	}
	else // bigger than 1TB
	{	dSize = ((double)size) / ((double)llTeraB);
		snprintf(text, max, "%.2f TB", (float) dSize);
	}
	
	return text;
}

int mkdir_recursive(char *path)
{
	char buffer[PATH_MAX];
	struct stat64 statbuf;
	int len;
	int pos;

	len=strlen(path);

	for (pos=0; pos<=len; pos++)
	{
		if (path[pos]=='/' || path[pos]==0)
		{
			memset(buffer, 0, sizeof(buffer));
			memcpy(buffer, path, pos);
			if (stat64(buffer, &statbuf)==-1 && errno==ENOENT)
				mkdir(buffer, 0755);
		}
	}
	
	return 0;
}

int extract_dirpath(char *filepath, char *dirbuf, int dirbufsize)
{
	int i;

	snprintf(dirbuf, dirbufsize, "%s", filepath);
	for (i=0; (i<dirbufsize) && (dirbuf[i]!=0); i++);
	while ((i>=0) && (dirbuf[i]!='/'))
		dirbuf[i--]=0;
	if ((i>0) && (dirbuf[i]=='/'))
		dirbuf[i]=0;
	return 0;
}

char *get_objtype_name(int objtype)
{
	switch (objtype)
	{
		case OBJTYPE_DIR: // case-1: directory
			return ("DIR     ");
		case OBJTYPE_SYMLINK: // case-2: symlink
			return ("SYMLINK ");
		case OBJTYPE_REGFILE: // case-3: regular file
			return ("REGFILE ");
		case OBJTYPE_HARDLINK: // case-4: hard link
			return ("HARDLINK");
		case OBJTYPE_CHARDEV: // case-5: char-dev
			return ("CHARDEV ");
		case OBJTYPE_BLOCKDEV: // case-6: block-dev
			return ("BLOCKDEV");
		case OBJTYPE_FIFO: // case-7: fifo
			return ("FIFO    ");
		case OBJTYPE_SOCKET: // case-8: socket
			return ("SOCKET  ");
		default: // case-8: unknown inode item
			return ("UNKNOWN ");
	}
}

int is_dir_empty(char *path)
{
	char fullpath[PATH_MAX];
	struct stat64 statbuf;
	struct dirent *dir;
	DIR *dirdesc;
	
	dirdesc = opendir(path);
	if (!dirdesc)
	{       sysprintf("cannot open directory %s\n", path);
		return -1;
	}
	
	while ((dir = readdir(dirdesc)) != NULL)
	{
		concatenate_paths(fullpath, sizeof(fullpath), path, dir->d_name);
		if (lstat64(fullpath, &statbuf)!=0)
		{	sysprintf ("cannot stat %s\n", fullpath);
			closedir(dirdesc);
			return -1;
		}
		if (strcmp(dir->d_name,".")!=0 && strcmp(dir->d_name,"..")!=0)
		{
			closedir(dirdesc);
			return 1; // an item was found
		}		
	}

	closedir(dirdesc);
	return 0;
}

// generate a non-null random u32
u32 generate_random_u32_id(void)
{
	struct timeval now;
	u32 archid;

	memset(&now, 0, sizeof(struct timeval));
	do
	{	gettimeofday(&now, NULL);
		archid=((u32)now.tv_sec)^((u32)now.tv_usec);
	} while (archid==0);
	return archid;
}

u32 fletcher32(u8 *data, u32 len)
{
	u32 sum1 = 0xffff, sum2 = 0xffff;
	
	while (len)
	{
		unsigned tlen = len > 360 ? 360 : len;
		len -= tlen;
		do {
			sum1 += *data++;
			sum2 += sum1;
		} while (--tlen);
		sum1 = (sum1 & 0xffff) + (sum1 >> 16);
		sum2 = (sum2 & 0xffff) + (sum2 >> 16);
	}
	// Second reduction step to reduce sums to 16 bits
	sum1 = (sum1 & 0xffff) + (sum1 >> 16);
	sum2 = (sum2 & 0xffff) + (sum2 >> 16);
	return sum2 << 16 | sum1;
}

int regfile_exists(char *filepath)
{
	struct stat64 st;
	int res;

	res=stat64(filepath, &st);
	if ((res==0) && S_ISREG(st.st_mode))
		return true; // file exists
	if (res==-1 && errno==ENOENT)
		return false; // does not exist
	return -1; // don't know
}

int exec_command(char *command, int cmdbufsize, char *format, ...)
{
	const int max_argv=32;
	char temp[2048];
	int pfildes1[2];
	int pfildes2[2];
	int status;
	int exitst;
	int mystdout;
	int mystderr;
	char delims[]=" ";
	char argv1[max_argv][1024];
	char *argv2[max_argv];
	char *saveptr;
	char *result;
	va_list ap;
	int cmdpid;
	int pid;
	int i;
	
	// init
	for(i=0; i<max_argv; i++)
		memset(argv1[i], 0, sizeof(argv1[i]));
	
	// format the string
	memset(command, 0, cmdbufsize);
	va_start(ap, format);
	vsnprintf(command, cmdbufsize, format, ap);
	va_end(ap);

	// parse the command
	memset(argv1, 0, sizeof(argv1));
	memset(argv2, 0, sizeof(argv2));
	snprintf(temp, sizeof(temp), "%s", command);
	
	result=strtok_r(temp, delims, &saveptr);
	for(i=0; (result != NULL) && (i<max_argv); i++)
	{
		snprintf(argv1[i], sizeof(argv1[i]), "%s", result);
		argv2[i]=argv1[i];
		msgprintf(MSG_DEBUG1, "argv1[%d]=[%s]\n", i, argv1[i]);
		result = strtok_r(NULL, delims, &saveptr);
	}
	
	// execute the command
	if ((pipe(pfildes1)==-1) || (pipe(pfildes2)==-1))
	{	errprintf("pipe() failed\n");
		return -1;
	}
	
	msgprintf(MSG_VERB1, "executing [%s]...\n", command);
	
	if ((pid = fork()) == -1)
	{	errprintf("fork() failed\n");
		return -1;
	}
	
	if (pid == 0) // child process --> command to be executed
	{	close(pfildes1[0]); // close read end of pipe
		close(pfildes2[0]); // close read end of pipe
		dup2(pfildes1[1],1); // make 1 same as write-to end of pipe
		close(pfildes1[1]); // close excess fildes
		dup2(pfildes2[1],2); // make 1 same as write-to end of pipe
		close(pfildes2[1]); // close excess fildes
		execvp(argv2[0], argv2);
		errprintf("execvp(%s) failed\n", argv2[0]); // still around? exec failed
		return -1;
	}
	else // parent --> fsarchiver
	{	mystdout=pfildes1[0];
		mystderr=pfildes2[0];
		cmdpid=pid;
		close(pfildes1[1]); // close write end of pipe
		close(pfildes2[1]); // close write end of pipe
		waitpid(cmdpid, &status, 0);
		exitst=WEXITSTATUS(status);
		msgprintf(MSG_VERB1, "command [%s] returned %d\n", command, exitst);
		memset(temp, 0, sizeof(temp));
		if ((exitst!=0) && (read(mystdout, temp, sizeof(temp))>0))
			msgprintf(MSG_DEBUG1, "\n----stdout----\n%s\n----stdout----\n\n", temp);
		memset(temp, 0, sizeof(temp));
		if ((exitst!=0) && (read(mystderr, temp, sizeof(temp))>0))
			msgprintf(MSG_DEBUG1, "\n----stderr----\n%s\n----stderr----\n\n", temp);
		return exitst;
	}
}

int generate_random_tmpdir(char *buffer, int bufsize)
{
        struct tm tbreak;
        time_t abstime;
	struct timeval tv;

	gettimeofday(&tv, NULL);
	abstime=tv.tv_sec;
        localtime_r(&abstime, &tbreak);

	snprintf(buffer, bufsize, "/tmp/fsarchiver/fsa-%.4d%.2d%.2d-%.2d%.2d%.2d",
		tbreak.tm_year+1900, tbreak.tm_mon+1, tbreak.tm_mday,
 		tbreak.tm_hour, tbreak.tm_min, tbreak.tm_sec);
	return 0;
}

int check_filesystem_support(char *filesystem)
{
	char delims[]=" \t\n";
	char line[1024];
	char *saveptr;
	char *result;
	int ret=-1;
	FILE *f;
	int i;
	
	if ((f=fopen("/proc/filesystems","rb"))==NULL)
	{	sysprintf("Cannot open /proc/filesystems\n");
		return -1;
	}
	
	while(!feof(f))
	{
		if (stream_readline(f, line, 1024)>1)
		{
			result=strtok_r(line, delims, &saveptr);
			for(i=0; result != NULL; i++)
			{
				if (strncmp(result, filesystem, sizeof(line))==0)
				{	ret=0;
					msgprintf(3, "filesystem [%s] seems to be supported by the kernel\n", filesystem);
					goto check_filesystem_support_close;
				}
				result = strtok_r(NULL, delims, &saveptr);
			}
		}
	}
	msgprintf(MSG_VERB2, "filesystem [%s] does not seems to be supported by the kernel\n", filesystem);
	
check_filesystem_support_close:
	fclose(f);
	return ret;
}

// returns true if magic is a valid magic-string
int is_magic_valid(char *magic)
{
	int i;
	for (i=0; valid_magic[i]!=NULL; i++)
		if (memcmp(magic, valid_magic[i], FSA_SIZEOF_MAGIC)==0)
			return true;
	return false;
}

// just copies the path if it has the right extension or add the extension
int path_force_extension(char *buf, int bufsize, char *origpath, char *ext)
{
	int oldlen, extlen;
	
	if (!buf || !origpath || !ext)
	{	errprintf("a parameter is null\n");
		return -1;
	}
	
	oldlen=strlen(origpath);
	extlen=strlen(ext);
	
	if ((oldlen < extlen) || memcmp(origpath+oldlen-extlen, ext, extlen)!=0)
		snprintf(buf, bufsize, "%s%s", origpath, ext);
	else // the extension has been found at the end of origpath
		snprintf(buf, bufsize, "%s", origpath);
	
	return 0;
}

// convert a binary md5 to an hexadecimal format
char *format_md5(char *buf, int maxbuf, u8 *md5bin)
{
	char temp[256];
	int i;
	
	memset(buf, 0, maxbuf);
	for (i=0; i<16; i++)
	{
		snprintf(temp, sizeof(temp), "%.2x", md5bin[i]);
		bufcat(buf, temp, maxbuf);
	}

	return buf;
}

// replacement for strncat() where we can't specify the destbuf size
char *bufcat(char *dest, const char *src, int n)
{
	int i, pos;
	
	for (i=strlen(dest), pos=0; (i < n-1) && (src[pos] != 0) ; i++, pos++)
		dest[i] = src[pos];
	dest[i]=0;
	return dest;
}
