/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>

#include "dico.h"
#include "common.h"

struct s_dico *dico_alloc()
{
	struct s_dico *d;
	if ((d=malloc(sizeof(struct s_dico)))==NULL)
		return NULL;
	d->head=NULL;
	return d;
}

int dico_destroy(struct s_dico *d)
{
	struct s_item *item, *next;
	
	if (d==NULL)
		return -1;
	
	item=d->head;
	while (item!=NULL)
	{
		next=item->next;
		free(item->data);
		free(item);
		item=next;
	}
	
	free(d);
	
	return 0;
}

int dico_add_data(struct s_dico *d, u8 section, u64 key, const void *data, u16 size)
{
	return dico_add_internal(d, section, key, data, size, DICTYPE_DATA);
}

// add an item to the dico, fails if an item with that (section,key) already exists
int dico_add_internal(struct s_dico *d, u8 section, u64 key, const void *data, u16 size, int type)
{
	struct s_item *item, *lnew, *last;
	
	if (size<1 || size>DICO_MAX_DATASIZE)
		return -1;
	
	if (d==NULL)
	{	errprintf("invalid dictionnary\n");
		return -2;
	}
	
	// allocate object
	lnew=malloc(sizeof(struct s_item));
	if (!lnew)
	{	errprintf("malloc(%d) failed: out of memory\n", sizeof(struct s_item));
		return -DICO_EMEM;
	}
	memset(lnew, 0, sizeof(struct s_item));

	// go to the end of the item and check for duplicates
	if (d->head==NULL) // item is empty
	{	
		d->head=lnew;
	}
	else // item is not empty
	{	
		for (item=d->head; item!=NULL; item=item->next)
		{	last=item;
			if (item->section==section && item->key==key)
				return -DICO_EDUPLICATE;
		}
		last->next=lnew;
	}
	
	// copy key
	lnew->key=key;
	lnew->section=section;
	
	// allocate memory for data
	lnew->data=malloc(size);
	if (!lnew->data)
		return -DICO_EMEM; // out of memory
	
	// copy data
	memcpy(lnew->data, data, size);
	lnew->size=size;
	
	// type of data
	lnew->type=type;
	
	return 0;
}

int dico_get_data(struct s_dico *d, u8 section, u64 key, void *data, u16 maxsize, u16 *size)
{
	return dico_get_internal(d, section, key, data, maxsize, size);
}

int dico_get_internal(struct s_dico *d, u8 section, u64 key, void *data, u16 maxsize, u16 *size)
{
	struct s_item *item;
	
	if (size!=NULL)
		*size=0;
	
	if (d==NULL)
	{	errprintf("invalid dictionnary: d==null\n");
		return -1;
	}
	
	if (d->head==NULL)
	{	msgprintf(MSG_DEBUG1, "dico is empty\n");
		return -2;
	}
	
	if (maxsize<1 || maxsize>DICO_MAX_DATASIZE)
		return -DICO_EBADSIZE;
	
	for (item=d->head; item!=NULL; item=item->next)
	{	
		if ((item!=NULL) && (item->key==key && item->section==section))
		{
			if (item->size > maxsize) // item is too big
				return -3;
			memcpy(data, item->data, item->size);
			if (size!=NULL)
				*size=item->size;
			return 0;
		}
	}
	
	return -4; // not found
}

int dico_count_one_section(struct s_dico *d, u8 section)
{
	struct s_item *item;
	int count;
	
	if (d==NULL)
	{	errprintf("invalid dictionnary: d==null\n");
		return -DICO_EINVAL;
	}
	
	count=0;
	for (item=d->head; item!=NULL; item=item->next)
		if (item->section==section)
			count++;
	
	return count;
}

int dico_count_all_sections(struct s_dico *d)
{
	struct s_item *item;
	int count;
	
	if (d==NULL)
	{	errprintf("invalid dictionnary: d==null\n");
		return -DICO_EINVAL;
	}
	
	count=0;
	for (item=d->head; item!=NULL; item=item->next)
		count++;
	
	return count;
}

int dico_add_u32(struct s_dico *d, u8 section, u64 key, u32 data)
{
	u32 ledata;
	ledata=cpu_to_le32(data);
	return dico_add_internal(d, section, key, &ledata, sizeof(ledata), DICTYPE_U32);
}

int dico_add_s32(struct s_dico *d, u8 section, u64 key, s32 data)
{
	s32 ledata;
	ledata=cpu_to_le32(data);
	return dico_add_internal(d, section, key, &ledata, sizeof(ledata), DICTYPE_S32);
}

int dico_add_u64(struct s_dico *d, u8 section, u64 key, u64 data)
{
	u64 ledata;
	ledata=cpu_to_le64(data);
	return dico_add_internal(d, section, key, &ledata, sizeof(ledata), DICTYPE_U64);
}

int dico_add_s64(struct s_dico *d, u8 section, u64 key, s64 data)
{
	s64 ledata;
	ledata=cpu_to_le64(data);
	return dico_add_internal(d, section, key, &ledata, sizeof(ledata), DICTYPE_S64);
}

int dico_add_string(struct s_dico *d, u8 section, u64 key, const char *szstring)
{
	u16 len;
	len=strlen(szstring);
	return dico_add_internal(d, section, key, szstring, len+1, DICTYPE_STRING);
}

int dico_get_u32(struct s_dico *d, u8 section, u64 key, u32 *data)
{
	u32 ledata;
	u16 size;
	
	*data=0;
	if (dico_get_data(d, section, key, &ledata, sizeof(ledata), &size)!=0)
		return -1;
	*data=le32_to_cpu(ledata);
	return 0;
}

int dico_get_s32(struct s_dico *d, u8 section, u64 key, s32 *data)
{
	s32 ledata;
	u16 size;
	
	*data=0;
	if (dico_get_data(d, section, key, &ledata, sizeof(ledata), &size)!=0)
		return -1;
	*data=le32_to_cpu(ledata);
	return 0;
}

int dico_get_u64(struct s_dico *d, u8 section, u64 key, u64 *data)
{
	u64 ledata;
	u16 size;
	
	*data=0;
	if (dico_get_data(d, section, key, &ledata, sizeof(ledata), &size)!=0)
		return -1;
	*data=le64_to_cpu(ledata);
	return 0;
}

int dico_get_s64(struct s_dico *d, u8 section, u64 key, s64 *data)
{
	s64 ledata;
	u16 size;
	
	*data=0;
	if (dico_get_data(d, section, key, &ledata, sizeof(ledata), &size)!=0)
		return -1;
	*data=le64_to_cpu(ledata);
	return 0;
}

int dico_get_string(struct s_dico *d, u8 section, u64 key, char *buffer, u16 bufsize)
{
	u16 size;
	memset(buffer, 0, bufsize);
	return dico_get_data(d, section, key, buffer, bufsize, &size);
}

int dico_show(struct s_dico *d, u8 section, char *debugtxt)
{
	char buffer[2048];
	char text[2048];
	struct s_item *item;
	void *dat;
	
	msgprintf(MSG_FORCE, "\n-----------------debug-dico-begin(%s)---------------\n", debugtxt);
	
	if (d->head)
	{
		for (item=d->head; item!=NULL; item=item->next)
		{
			if ((item->data!=NULL) && (item->size>0) && (item->section==section))
			{
				snprintf(buffer, sizeof(buffer), "key=[%ld], sizeof(data)=[%d], ", (long)item->key, (int)item->size);
				dat=item->data;
				
				switch (item->type)
				{
					case DICTYPE_U8:
						snprintf(text, sizeof(text), "type=u8, data=[%ld]", (long)*((u8*)dat));
						break;
					case DICTYPE_U16:
						snprintf(text, sizeof(text), "type=u16, data=[%ld]", (long)*((u16*)dat));
						break;
					case DICTYPE_U32:
						snprintf(text, sizeof(text), "type=u32, data=[%ld]", (long)*((u32*)dat));
						break;
					case DICTYPE_U64:
						snprintf(text, sizeof(text), "type=u64, data=[%ld]", (long)*((u64*)dat));
						break;
					case DICTYPE_S8:
						snprintf(text, sizeof(text), "type=u8, data=[%ld]", (long)*((s8*)dat));
						break;
					case DICTYPE_S16:
						snprintf(text, sizeof(text), "type=u16, data=[%ld]", (long)*((s16*)dat));
						break;
					case DICTYPE_S32:
						snprintf(text, sizeof(text), "type=u32, data=[%ld]", (long)*((s32*)dat));
						break;
					case DICTYPE_S64:
						snprintf(text, sizeof(text), "type=u64, data=[%ld]", (long)*((s64*)dat));
						break;
					case DICTYPE_STRING:
						snprintf(text, sizeof(text), "type=str, data=[%s]", (char*)dat);
						break;
					case DICTYPE_DATA:
						snprintf(text, sizeof(text), "type=dat");
						/*memset(text, 0, sizeof(text));
						for (i=0; i<item->size; i++)
							printf("%.2x", dat[i]);
						msgprintf(1, "]=[%s]\n", dat);*/
						break;
					default:
						snprintf(text, sizeof(text), "type=unknown");
						break;
				}
				bufcat(buffer, text, sizeof(buffer));
				msgprintf(1, "%s\n", buffer);
			}
		}
	}
	else
	{
		msgprintf(MSG_FORCE, "dico is empty\n");
	}
	
	msgprintf(MSG_FORCE, "-----------------debug-dico-end(%s)------------------\n\n", debugtxt);
	return 0;
}
