/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <stdlib.h>
#include <utime.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <linux/limits.h>
#include <attr/xattr.h>
#include <sys/statvfs.h>

#include "dico.h"
#include "common.h"
#include "extract.h"
#include "archive.h"
#include "archinfo.h"
#include "md5.h"
#include "mount.h"
#include "fs_ext2.h"
#include "fs_reiserfs.h"
#include "fs_reiser4.h"
#include "fs_btrfs.h"
#include "fs_xfs.h"
#include "fs_jfs.h"
#include "thread_comp.h"
#include "thread_archio.h"
#include "syncthread.h"

int restore_obj_symlink(struct s_archive *ai, char *fullpath, char *destdir, struct s_dico *d)
{
	char buffer[PATH_MAX];

	if (dico_get_string(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_SYMLINK, buffer, PATH_MAX)<0)
	{	errprintf("Cannot read field=symlink for file=[%s]\n", fullpath);
		return -1;
	}

	if (symlink(buffer, fullpath)<0)
	{	sysprintf("symlink(%s, %s) failed\n", buffer, fullpath);
		return -1;
	}
	return 0;
}

int restore_obj_hardlink(struct s_archive *ai, char *fullpath, char *destdir, struct s_dico *d)
{
	char buffer[PATH_MAX];
	char regfile[PATH_MAX];

	if (dico_get_string(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_HARDLINK, buffer, PATH_MAX)<0)
		return -1;
	concatenate_paths(regfile, PATH_MAX, destdir, buffer);
	if (link(regfile, fullpath)!=0)
		return -2;

	return 0;
}

int restore_obj_devfile(struct s_archive *ai, char *fullpath, char *destdir, struct s_dico *d)
{
	dev_t dev;
	u64 temp;
	u32 mode;

	if (dico_get_u64(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_RDEV, &temp)!=0)
		return -1;
	if (dico_get_u32(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_MODE, &mode)!=0)
		return -2;
	dev=temp;
	if (mknod(fullpath, mode, dev)!=0)
		return -3;
	return 0;
}

int restore_obj_directory(struct s_archive *ai, char *fullpath, char *destdir, struct s_dico *d)
{
	mkdir_recursive(fullpath);
	return 0;
}

int restore_obj_regfile(struct s_archive *ai, char *relpath, char *fullpath, char *destdir, struct s_dico *d)
{
	struct s_dico *footerdico=NULL;
	char magic[FSA_SIZEOF_MAGIC+1];
	struct statvfs64 statvfsbuf;
	struct s_blockinfo blkinfo;
	struct md5_ctx md5ctx;
	char textbuf[128];
	u8 md5sumcalc[16];
	u8 md5sumorig[16];
	int errors;
	char text[256];
	u64 filesize;
	u64 filepos;
	int ret=0;
	s64 lres;
	int fd;
	
	// init
	errors=0;
	memset(&blkinfo, 0, sizeof(blkinfo));
	memset(magic, 0, sizeof(magic));
	md5_init_ctx(&md5ctx);
	
	if (dico_get_u64(d, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_SIZE, &filesize)!=0)
	{	errprintf("Cannot read filesize DISKITEMKEY_SIZE from archive for file=[%s]\n", relpath);
		return -1;
	}
	
	fd=open64(fullpath, O_RDWR|O_CREAT|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (fd<0)
	{	sysprintf("Cannot open %s for writing\n", relpath);
		return -2;
	}
	
	msgprintf(MSG_DEBUG1, "restore_obj_regfile(file=%s, size=%lld)\n", relpath, (long long)filesize);
	for (filepos=0; (filepos < filesize) && (get_interrupted()==false); filepos+=blkinfo.blkrealsize)
	{
		if ((lres=queue_dequeue_block(&g_queue, &blkinfo))<=0)
		{	errprintf("queue_dequeue_block()=%ld=%s for file(%s) failed\n", (long)lres, qerr(lres), relpath);
			errors++;
			goto restore_obj_error;
		}
		
		msgprintf(MSG_DEBUG2, "----> filepos=%lld, blkinfo.blkoffset=%lld, blkinfo.blkrealsize=%lld\n", 
			(long long)filepos, (long long)blkinfo.blkoffset, (long long)blkinfo.blkrealsize);
		
		if (blkinfo.blkoffset!=filepos)
		{	errprintf("file offset do not match for file(%s) failed: filepos=%lld, blkinfo.blkoffset=%lld, blkinfo.blkrealsize=%lld\n", 
				relpath, (long long)filepos, (long long)blkinfo.blkoffset, (long long)blkinfo.blkrealsize);
			errors++;
			goto restore_obj_error;
		}
		
		md5_process_bytes(blkinfo.blkdata, blkinfo.blkrealsize, &md5ctx);
		if ((lres=write(fd, blkinfo.blkdata, blkinfo.blkrealsize))!=blkinfo.blkrealsize) // error
		{
			if ((lres>0) && (lres < blkinfo.blkrealsize)) // probably "no space left"
			{
				if (fstatvfs64(fd, &statvfsbuf)!=0)
				{	sysprintf("fstatvfs64(fd=%d) failed\n", fd);
					ret=-5;
					goto restore_obj_error;
				}
				
				u64 freebytes = statvfsbuf.f_bfree * statvfsbuf.f_bsize;
				errprintf("Can't write all data for file [%s]: space on device is %s\n", relpath, 
					format_size(freebytes, textbuf, sizeof(textbuf), 'h'));
				ret=-5;
				goto restore_obj_error;
			}
			else // another error
			{
				sysprintf("cannot extract %s, curblocksize=%ld\n", relpath, (long)blkinfo.blkrealsize);
				ret=-5;
				goto restore_obj_error;
			}
		}
		free(blkinfo.blkdata);
	}
	
	if (get_interrupted()==true)
	{	errprintf("operation has been interrupted\n");
		ret=-1;
		goto restore_obj_error;
	}
		
	// check the whole file md5 checksum
	md5_finish_ctx(&md5ctx, md5sumcalc);
	msgprintf(MSG_DEBUG1, "--> finished loop for file=%s, size=%lld, md5sumcalc=[%s]\n", relpath, 
			(long long)filesize, format_md5(text, sizeof(text), md5sumcalc));
	
	if (queue_dequeue_header(&g_queue, &footerdico, magic)<=0)
	{	errprintf("queue_dequeue_header() failed: cannot read footer dico\n");
		errors++;
		goto restore_obj_error;
	}
	
	if (memcmp(magic, FSA_MAGIC_FILF, FSA_SIZEOF_MAGIC)!=0)
	{	errprintf("header is not what we expected: found=[%s] and expected=[%s]\n", magic, FSA_MAGIC_FILF);
		goto restore_obj_error;
	}
	
	if (dico_get_data(footerdico, 0, BLOCKFOOTITEMKEY_MD5SUM, md5sumorig, 16, NULL))
	{	errprintf("cannot get md5sum from file footer for file=[%s]\n", relpath);
		dico_show(footerdico, 0, "footerdico");
		ret=-7;
		goto restore_obj_error;
	}
	
	if (memcmp(md5sumcalc, md5sumorig, 16)!=0)
	{	errprintf("cannot restore file %s, file is corrupt\n");
		errors++;
	}
	
restore_obj_error:
	if (footerdico!=NULL)
		dico_destroy(footerdico);
	close(fd);
	return ret;
}

int restore_object(struct s_archive *ai, int *errors, char *destdir, struct s_dico *dicoattr, int fsid)
{
	char relpath[PATH_MAX];
	char fullpath[PATH_MAX];
	char parentdir[PATH_MAX];
	char xattrname[1024];
	char xattrvalue[1024];
	u16 xattrdatasize;
	u32 mode, uid, gid;
	u64 atime, mtime;
	struct utimbuf times;
	u32 objtype;
	int xattrdicsize;
	int res;
	int i;
	
	// init
	*errors=0;
	
	if (dico_get_data(dicoattr, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_PATH, relpath, sizeof(relpath), NULL)!=0)
		return -1;
	if (dico_get_u32(dicoattr, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_OBJTYPE, &objtype)!=0)
		return -2;
	concatenate_paths(fullpath, sizeof(fullpath), destdir, relpath);
	
	msgprintf(MSG_VERB1, "-[%.2d][%s] %s\n", fsid, get_objtype_name(objtype), relpath);
	
	// create parent directory first
	extract_dirpath(fullpath, parentdir, sizeof(parentdir));
	mkdir_recursive(parentdir);
	
	switch (objtype)
	{
		case OBJTYPE_DIR:
			msgprintf(MSG_DEBUG1, "objtype=OBJTYPE_DIR, path=[%s]\n", relpath);
			res=restore_obj_directory(ai, fullpath, destdir, dicoattr);
			break;
		case OBJTYPE_SYMLINK:
			msgprintf(MSG_DEBUG1, "objtype=OBJTYPE_SYMLINK, path=[%s]\n", relpath);
			res=restore_obj_symlink(ai, fullpath, destdir, dicoattr);
			break;
		case OBJTYPE_REGFILE:
			msgprintf(MSG_DEBUG1, "objtype=OBJTYPE_REGFILE, path=[%s]\n", relpath);
			res=restore_obj_regfile(ai, relpath, fullpath, destdir, dicoattr);
			if (res<0)
			{	errprintf("restore_obj_regfile(%s) failed with res=%d\n", relpath, res);
			}
			else if (res>0) // corruptions found
			{	(*errors)++;
				res=0; // continue anyway
			}
			break;
		case OBJTYPE_HARDLINK:
			msgprintf(MSG_DEBUG1, "objtype=OBJTYPE_HARDLINK, path=[%s]\n", relpath);
			res=restore_obj_hardlink(ai, fullpath, destdir, dicoattr);
			break;
		case OBJTYPE_CHARDEV:
			msgprintf(MSG_DEBUG1, "objtype=OBJTYPE_CHARDEV, path=[%s]\n", relpath);
			res=restore_obj_devfile(ai, fullpath, destdir, dicoattr);
			break;
		case OBJTYPE_BLOCKDEV:
			msgprintf(MSG_DEBUG1, "objtype=OBJTYPE_BLOCKDEV, path=[%s]\n", relpath);
			res=restore_obj_devfile(ai, fullpath, destdir, dicoattr);
			break;
		case OBJTYPE_FIFO:
			msgprintf(MSG_DEBUG1, "objtype=OBJTYPE_FIFO, path=[%s]\n", relpath);
			res=restore_obj_devfile(ai, fullpath, destdir, dicoattr);
			break;
		case OBJTYPE_SOCKET:
			msgprintf(MSG_DEBUG1, "objtype=OBJTYPE_SOCKET, path=[%s]\n", relpath);
			res=restore_obj_devfile(ai, fullpath, destdir, dicoattr);
			break;
		default:
			errprintf("Unknown objtype %d\n", objtype);
			return -3;
	}
	
	// ---- restore standard attributes (permissions, owner, ...)
	if (res!=0) // value returned by restore_obj_xxx()
	{	errprintf("Restoring file=[%s], objtype=[%s] failed\n", fullpath, get_objtype_name(objtype));
		return -1;
	}
	if (dico_get_u32(dicoattr, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_MODE, &mode)!=0)
		return -4;
	if (dico_get_u32(dicoattr, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_UID, &uid)!=0)
		return -5;
	if (dico_get_u32(dicoattr, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_GID, &gid)!=0)
		return -6;
	if (dico_get_u64(dicoattr, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_ATIME, &atime)!=0)
		return -7;
	if (dico_get_u64(dicoattr, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_MTIME, &mtime)!=0)
		return -8;
	times.actime=(time_t)atime;
	times.modtime=(time_t)mtime;
	
	if (lchown(fullpath, (uid_t)uid, (gid_t)gid)!=0)
	{	sysprintf("Cannot lchown(%s) which is %s\n", fullpath, get_objtype_name(objtype));
		return -10;
	}
	
	if (objtype!=OBJTYPE_SYMLINK)
	{
		if (chmod(fullpath, (mode_t)mode)!=0)
			return -9;
	}
	
	// set the values for atime/mtime
	struct timeval tv[2];
	tv[0].tv_usec=0;
	tv[0].tv_sec=atime;
	tv[1].tv_usec=0;
	tv[1].tv_sec=mtime;
	if (lutimes(fullpath, tv)!=0)
		return -11;
	
	// ---- restore extended attributes
	xattrdicsize=dico_count_one_section(dicoattr, DICO_OBJ_SECTION_XATTR);
	
	for (i=0; i < xattrdicsize; i+=2)
	{
		if (dico_get_string(dicoattr, DICO_OBJ_SECTION_XATTR, (u64)(i+0), xattrname, sizeof(xattrname))!=0)
		{	errprintf("Cannot retrieve the name of an xattr for file %s\n", relpath);
			dico_show(dicoattr, DICO_OBJ_SECTION_XATTR, "xattr");
			return -1;
		}
		memset(xattrvalue, 0, sizeof(xattrvalue));
		if (dico_get_data(dicoattr, DICO_OBJ_SECTION_XATTR, (u64)(i+1), xattrvalue, sizeof(xattrvalue), &xattrdatasize)!=0)
		{	errprintf("Cannot retrieve the value of an xattr for file %s\n", relpath);
			return -1;
		}
		
		res=lsetxattr(fullpath, xattrname, xattrvalue, xattrdatasize, 0);
		if (res!=0)
		{ sysprintf("lsetxattr(%s,%s,%s) failed\n", relpath, xattrname, xattrvalue); }
		else
		{ msgprintf(MSG_VERB2, "            lsetxattr(%s, %s, %s)=%d\n", relpath, xattrname, xattrvalue, res); }
	}
	
	return 0;
}

int extract_read_objects(struct s_archive *ai, int *errors, char *destdir, int fsid)
{
	char magic[FSA_SIZEOF_MAGIC+1];
	struct s_dico *dicoattr=NULL;
	int headerisend;
	int headerisobj;
	u64 checkfsid;
	int curerr;
	int type;
	int res;
	
	// init
	memset(magic, 0, sizeof(magic));
	*errors=0;
	
	do
	{	// skip the garbage (just ignore everything until the next FSA_MAGIC_OBJT)
		// in case the archive is corrupt and random data has been added / removed in the archive
		do
		{	if (queue_check_next_item(&g_queue, &type, magic)!=0)
			{	errprintf("queue_check_next_item() failed: cannot read object from archive\n");
				return -1;
			}
			
			headerisobj=(memcmp(magic, FSA_MAGIC_OBJT, FSA_SIZEOF_MAGIC)==0);
			headerisend=(memcmp(magic, FSA_MAGIC_DATF, FSA_SIZEOF_MAGIC)==0);
			
			if (headerisobj!=true && headerisend!=true) // did not find expected header: skip garbage in archive
			{	
				msgprintf(1, "unexpected header found in archive, skipping it: type=%d, magic=[%s]\n", 
					type, (type==QITEM_TYPE_DICO)?(magic):"-block-");
				if (queue_destroy_first_item(&g_queue)!=0)
				{	errprintf("queue_destroy_first_item() failed: cannot read object from archive\n");
					return -1;
				}
			}
		} while ((headerisobj!=true) && (headerisend!=true));
		
		if (headerisobj==true) // if it's an object header
		{
			// read object header from archive
			if (queue_dequeue_header(&g_queue, &dicoattr, magic)<=0)
			{	errprintf("queue_dequeue_header() failed\n");
				return -1;
			}
			
			// check fsid of that object
			if (dico_get_u64(dicoattr, 0, DISKITEMKEY_FSID, &checkfsid)!=0)
			{	errprintf("cannot find DISKITEMKEY_FSID in  objectheader\n");
				return -1;
			}
			
			if (checkfsid==fsid) // if filesystem-id is correct
			{
				if ((res=restore_object(ai, &curerr, destdir, dicoattr, fsid))!=0)
				{	errprintf("restore_object() failed with res=%d\n", res);
					dico_destroy(dicoattr);
					return -1;
				}
				else if (curerr>0)
				{
					*errors+=curerr;
				}
			}
			else // wrong filesystem-id
			{	errprintf("restore_object(): object has a wrong filesystem id: found=[%d], expected=[%d]\n", checkfsid, fsid);
				(*errors)++;
			}
			
			dico_destroy(dicoattr);
		}
	} while ((headerisend!=true) && (get_abort()==false));
	
	return 0;
}

int mkfs(struct s_dico *d, char *partition, char *filesystem)
{
	// create filesystem
	if ((strcmp(filesystem, "ext2")==0) || (strcmp(filesystem, "ext3")==0) || (strcmp(filesystem, "ext4")==0))
		return ext2_mkfs(d, partition, filesystem);
	else if (strcmp(filesystem, "reiserfs")==0)
		return reiser_mkfs(d, partition, filesystem);
	else if (strcmp(filesystem, "reiser4")==0)
		return reiser4_mkfs(d, partition, filesystem);
	else if (strcmp(filesystem, "btrfs")==0)
		return btrfs_mkfs(d, partition, filesystem);
	else if (strcmp(filesystem, "xfs")==0)
		return xfs_mkfs(d, partition, filesystem);
	else if (strcmp(filesystem, "jfs")==0)
		return jfs_mkfs(d, partition, filesystem);
	else // unknown fs
		{ msgprintf(2, "unknown filesystem: [%s]\n", filesystem); return -1; }

	return 0;
}

int read_mainhead(struct s_archive *ai, struct s_dico **dicomainhead)
{
	char magic[FSA_SIZEOF_MAGIC+1];
	
	// init
	memset(magic, 0, sizeof(magic));
	
	if (!ai || !dicomainhead)
	{	errprintf("ai or dicomainhead is NULL\n");
		return -1;
	}
	
	if (queue_dequeue_header(&g_queue, dicomainhead, magic)<=0)
	{	errprintf("queue_dequeue_header() failed: cannot read main header\n");
		return -1;
	}
	
	if (memcmp(magic, FSA_MAGIC_MAIN, FSA_SIZEOF_MAGIC)!=0)
	{	errprintf("header is not what we expected: found=[%s] and expected=[%s]\n", magic, FSA_MAGIC_MAIN);
		return -1;
	}
	
	if (dico_get_u32(*dicomainhead, 0, MAINHEADKEY_ARCHTYPE, &ai->archtype)!=0)
	{	errprintf("cannot find MAINHEADKEY_ARCHTYPE in main-header\n");
		return -1;
	}
	
	if (ai->archtype==ARCHTYPE_FILESYSTEMS && dico_get_u64(*dicomainhead, 0, MAINHEADKEY_FSCOUNT, &ai->fscount)!=0)
	{	errprintf("cannot find MAINHEADKEY_FSCOUNT in main-header\n");
		return -1;
	}
	
	if (dico_get_u32(*dicomainhead, 0, MAINHEADKEY_ARCHIVEID, &ai->archid)!=0)
	{	errprintf("cannot find MAINHEADKEY_ARCHIVEID in main-header\n");
		return -1;
	}
	
	if (dico_get_data(*dicomainhead, 0, MAINHEADKEY_FILEFORMATVER, ai->filefmt, FSA_MAX_FILEFMTLEN, NULL)!=0)
	{	errprintf("cannot find MAINHEADKEY_FILEFORMATVER in main-header\n");
		return -1;
	}
	
	if (dico_get_data(*dicomainhead, 0, MAINHEADKEY_PROGVERCREAT, ai->creatver, FSA_MAX_PROGVERLEN, NULL)!=0)
	{	errprintf("cannot find MAINHEADKEY_PROGVERCREAT in main-header\n");
		return -1;
	}
	
	if (dico_get_data(*dicomainhead, 0, MAINHEADKEY_ARCHLABEL, ai->label, FSA_MAX_LABELLEN, NULL)!=0)
	{	errprintf("cannot find MAINHEADKEY_ARCHLABEL in main-header\n");
		return -1;
	}
	
	if (strcmp(ai->filefmt, FSA_FILEFORMAT)!=0)
	{
		errprintf("This archive is based on a different file format: [%s]. Cannot continue.\n", ai->filefmt);
		errprintf("It has been created with fsarchiver [%s], you should extrat the archive using that version.\n", ai->creatver);
		errprintf("The current version of the program is [%s], and it's based on format [%s]\n", FSA_VERSION, FSA_FILEFORMAT);
		return -1;
	}
	
	return 0;
}

// read headers and blocks from queue but do not write anything
int filesystem_skip(struct s_archive *ai)
{
	char magic[FSA_SIZEOF_MAGIC+1];
	struct s_dico *dicobegin=NULL;
	struct s_dico *dicoend=NULL;
	struct s_dico *dicoattr=NULL;
	struct s_dico *footerdico=NULL;
	struct s_blockinfo blkinfo;
	int headerisend;
	int headerisobj;
	u64 filesize;
	u64 filepos;
	u32 objtype;
	int type;
	int ret=0;
	
	// init
	memset(magic, 0, sizeof(magic));
	
	// ---- read filesystem-header from archive
	if (queue_dequeue_header(&g_queue, &dicobegin, magic)<=0)
	{	errprintf("queue_dequeue_header() failed: cannot read file system dico\n");
		return -1;
	}
	
	if (memcmp(magic, FSA_MAGIC_FSYB, FSA_SIZEOF_MAGIC)!=0)
	{	errprintf("header is not what we expected: found=[%s] and expected=[%s]\n", magic, FSA_MAGIC_FSYB);
		return -1;
	}
	dico_destroy(dicobegin);	

	// read all objects headers and blocks from queue
	do
	{	// skip the garbage (just ignore everything until the next FSA_MAGIC_OBJT)
		// in case the archive is corrupt and random data has been added / removed in the archive
		do
		{	if (queue_check_next_item(&g_queue, &type, magic)!=0)
			{	errprintf("queue_check_next_item() failed: cannot read object from archive\n");
				return -1;
			}
			
			headerisobj=(memcmp(magic, FSA_MAGIC_OBJT, FSA_SIZEOF_MAGIC)==0);
			headerisend=(memcmp(magic, FSA_MAGIC_DATF, FSA_SIZEOF_MAGIC)==0);
			
			if (headerisobj!=true && headerisend!=true) // did not find expected header: skip garbage in archive
			{	
				msgprintf(1, "unexpected header found in archive, skipping it: type=%d, magic=[%s]\n", 
					type, (type==QITEM_TYPE_DICO)?(magic):"-block-");
				if (queue_destroy_first_item(&g_queue)!=0)
				{	errprintf("queue_destroy_first_item() failed: cannot read object from archive\n");
					return -1;
				}
			}
		} while ((headerisobj!=true) && (headerisend!=true));
		
		if (headerisobj==true) // if it's an object header
		{
			// read object header from archive
			if (queue_dequeue_header(&g_queue, &dicoattr, magic)<=0)
			{	errprintf("queue_dequeue_header() failed\n");
				return -1;
			}
			
			if (dico_get_u32(dicoattr, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_OBJTYPE, &objtype)!=0)
			{	errprintf("dico_get_u32(DICO_OBJ_SECTION_STDATTR) failed\n");
				return -1;
			}
			
			if (objtype==OBJTYPE_REGFILE)
			{
				if (dico_get_u64(dicoattr, DICO_OBJ_SECTION_STDATTR, DISKITEMKEY_SIZE, &filesize)!=0)
				{	errprintf("Cannot read filesize from archive\n");
					return -1;
				}
				
				for (filepos=0; filepos<filesize; filepos+=blkinfo.blkrealsize)
				{
					if (queue_dequeue_block(&g_queue, &blkinfo)<=0)
						errprintf("queue_dequeue_block() failed\n");
					
					if (blkinfo.blkoffset!=filepos)
						errprintf("file offset do not match\n");
					
					free(blkinfo.blkdata);
				}
				
				if (queue_dequeue_header(&g_queue, &footerdico, magic)<=0)
					errprintf("queue_dequeue_header() failed: cannot read footer dico\n");
				dico_destroy(footerdico);
				
				if (memcmp(magic, FSA_MAGIC_FILF, FSA_SIZEOF_MAGIC)!=0)
					errprintf("header is not what we expected: found=[%s] and expected=[%s]\n", magic, FSA_MAGIC_FILF);
			}
			
			dico_destroy(dicoattr);
		}
	} while ((headerisend!=true) && (get_abort()==false));
	
	// read "end of file-system" header from archive
	if (queue_dequeue_header(&g_queue, &dicoend, magic)<=0)
	{	msgprintf(MSG_STACK, "queue_dequeue_header() failed\n");
		ret=-1;
		goto filesystem_skip_return;
	}
	dico_destroy(dicoend);
	if (memcmp(magic, FSA_MAGIC_DATF, FSA_SIZEOF_MAGIC)!=0)
	{	msgprintf(MSG_STACK, "header is not what we expected: found=[%s] and expected=[%s]\n", magic, FSA_MAGIC_DATF);
		goto filesystem_skip_return;
	}
	
filesystem_skip_return:
	return ret;
}

int filesystem_extract(struct s_archive *ai, struct s_dico *dicofs, struct s_dico *dicocmdline, int fsid)
{
	char filesystem[FSA_MAX_FSNAMELEN];
	char text[FSA_MAX_FSNAMELEN];
	char fsbuf[FSA_MAX_FSNAMELEN];
	char magic[FSA_SIZEOF_MAGIC+1];
	char partition[1024];
	char destfs[FSA_MAX_FSNAMELEN];
	struct s_dico *dicobegin=NULL;
	struct s_dico *dicoend=NULL;
	char mntbuf[PATH_MAX];
	u64 fsbytestotal;
	u64 fsbytesused;
	char optbuf[128];
	int readwrite;
	int errors=0;
	int ret=0;
	int res;
	
	// init
	memset(magic, 0, sizeof(magic));
	memset(destfs, 0, sizeof(destfs));
	memset(partition, 0, sizeof(partition));
	
	// read destination partition from dicocmdline
	if ((dico_get_string(dicocmdline, 0, FSINFOKEY_DEST, partition, sizeof(partition)))<0)
	{	errprintf("dico_get_string(FSINFOKEY_DEST) failed\n");
		return -1;
	}
	// read filesystem to create partition from dicocmdline (optionnal, may not exist: do not fail)
	dico_get_string(dicocmdline, 0, FSINFOKEY_MKFS, destfs, sizeof(destfs));
	
	// check the partition is not mounted
	res=mount_get_devinfo(partition, &readwrite, mntbuf, sizeof(mntbuf), optbuf, sizeof(optbuf), fsbuf, sizeof(fsbuf));
	if (res==0)
	{	errprintf("partition [%s] is mounted on [%s].\ncannot restore an archive to a partition "
			"which is mounted, unmount it first: umount %s\n", partition, mntbuf, mntbuf);
		return -1;
	}
	
	// ---- read filesystem-header from archive
	if (queue_dequeue_header(&g_queue, &dicobegin, magic)<=0)
	{	errprintf("queue_dequeue_header() failed: cannot read file system dico\n");
		return -1;
	}
	dico_destroy(dicobegin);
	
	if (memcmp(magic, FSA_MAGIC_FSYB, FSA_SIZEOF_MAGIC)!=0)
	{	errprintf("header is not what we expected: found=[%s] and expected=[%s]\n", magic, FSA_MAGIC_FSYB);
		return -1;
	}
	
	if ((dico_get_string(dicofs, 0, FSYSHEADKEY_FILESYSTEM, filesystem, sizeof(filesystem)))<0)
	{	errprintf("dico_get_string(FSYSHEADKEY_FILESYSTEM) failed\n");
		return -1;
	}
	
	if (dico_get_u64(dicofs, 0, FSYSHEADKEY_BYTESTOTAL, &fsbytestotal)!=0)
	{	errprintf("dico_get_string(FSYSHEADKEY_BYTESTOTAL) failed\n");
		return -1;
	}
	
	if (dico_get_u64(dicofs, 0, FSYSHEADKEY_BYTESUSED, &fsbytesused)!=0)
	{	errprintf("dico_get_string(FSYSHEADKEY_BYTESUSED) failed\n");
		return -1;
	}
	
	snprintf(fsbuf, sizeof(fsbuf), "%s", filesystem);
	msgprintf(MSG_VERB2, "filesystem=[%s]\n", filesystem);
	msgprintf(MSG_VERB2, "fsbytestotal=[%s]\n", format_size(fsbytestotal, text, sizeof(text), 'h'));
	msgprintf(MSG_VERB2, "fsbytesused=[%s]\n", format_size(fsbytesused, text, sizeof(text), 'h'));
	
	// ---- make the filesystem
	if (destfs[0]) // if a filesystem to use was specified: overwrite the default one
		snprintf(filesystem, sizeof(filesystem), "%s", destfs);
	
	// check if the destination filesystem is supported by the kernel
	if (check_filesystem_support(filesystem)!=0)
	{	errprintf("the filesystem %s does not seems to be supported by the current kernel (see /proc/filesystems)\n", filesystem);
		return -1;
	}
	
	if(mkfs(dicofs, partition, filesystem)!=0)
	{	errprintf("cannot format the filesystem %s on partition %s\n", filesystem, partition);
		return -1;
	}
	
	// ---- mount the new filesystem
	mkdir_recursive(mntbuf);
	generate_random_tmpdir(mntbuf, sizeof(mntbuf));
	mkdir_recursive(mntbuf);
	snprintf(fsbuf, sizeof(fsbuf), "%s", filesystem);
	if (mount_partition(partition, mntbuf, fsbuf, sizeof(fsbuf), 0)!=0)
	{	errprintf("partition [%s] cannot be mounted on %s. cannot continue.\n", partition, mntbuf);
		return -1;
	}
	
	if (extract_read_objects(ai, &errors, mntbuf, fsid)!=0)
	{	msgprintf(MSG_STACK, "extract_read_objects(%s) failed\n", mntbuf);
		return -1;
	}
	else if (errors>0)
	{	msgprintf(MSG_DEBUG1, "extract_read_objects(%s) worked with errors\n", mntbuf);
		ret=-1;
		goto filesystem_extract_umount;
	}
	
	// read "end of file-system" header from archive
	if (queue_dequeue_header(&g_queue, &dicoend, magic)<=0)
	{	errprintf("queue_dequeue_header() failed\n");
		ret=-1;
		goto filesystem_extract_umount;
	}
	dico_destroy(dicoend);
	
	if (memcmp(magic, FSA_MAGIC_DATF, FSA_SIZEOF_MAGIC)!=0)
	{	errprintf("header is not what we expected: found=[%s] and expected=[%s]\n", magic, FSA_MAGIC_DATF);
		goto filesystem_extract_umount;
	}
	
filesystem_extract_umount:
	if (unmount_partition(mntbuf)!=0)
	{	sysprintf("cannot umount %s\n", mntbuf);
		ret=-1;
	}
	return ret;
}

// convert a string such as "id=0,dest=/dev/sda1,fs=reiserfs" to a dico
int convert_string_to_dico(struct s_dico **d, int *retfsid, char *argv, int fscount)
{
	char name[128];
	char value[128];
	char *saveptr;
	char *result;
	char delims[]=",\t\n";
	struct stat64 st;
	int i, pos;
	
	// init 
	*retfsid=-1;
	
	// init
	if ((*d=dico_alloc())==NULL)
	{	errprintf("dico_alloc() failed\n");
		return -1;
	}
	
	result=strtok_r(argv, delims, &saveptr);
	while (result!=NULL)
	{
		memset(name, 0, sizeof(name));
		memset(value, 0, sizeof(value));
		
		for (i=0; (result[i]!=0) && (result[i]!='=') && (i<sizeof(name)); i++)
		{
			name[i]=result[i];
		}
		if (result[i++]!='=')
		{	errprintf("Incorrect syntax in the \"restfs\" command. Cannot find symbol '=' in [%s], expected something like [name1=val1,name2=val2]\n", result);
			return -1;
		}
		for (pos=0; (result[i]!=0) && (pos<sizeof(value)); pos++)
		{
			value[pos]=result[i++];
		}
		
		if (strcmp(name, "id")==0)
		{
			*retfsid=atoi(value);
			if (*retfsid<0 || *retfsid>fscount-1)
			{	errprintf("the filesystem id must be an integer between 0 and %d\n", fscount-1);
				return -1;
			}
			dico_add_u64(*d, 0, FSINFOKEY_FSID, *retfsid);
		}
		else if (strcmp(name, "dest")==0)
		{
			if ((stat64(value, &st)!=0) || (!S_ISBLK(st.st_mode)))
			{	errprintf("%s is not a valid block device\n", value);
				return -1;
			}
			dico_add_string(*d, 0, FSINFOKEY_DEST, value);
		}
		else if (strcmp(name, "mkfs")==0)
		{
			if (check_filesystem_support(value)!=0)
			{	errprintf("the filesystem %s does not seems to be supported by the current kernel (see /proc/filesystems)\n", value);
				return -1;
			}
			dico_add_string(*d, 0, FSINFOKEY_MKFS, value);
		}
		else
		{	errprintf("invalid key: [%s], expected 'id' or 'dest' or 'mkfs'\n", name);
			return -1;
		}
		result=strtok_r(NULL, delims, &saveptr);
	}
	
	return 0;
}

// convert an array of strings to an array of dicos
int convert_argv_to_dicos(struct s_dico *dicoargv[], char *cmdargv[], int fscount)
{
	struct s_dico *tmpdico=NULL;
	int fsid;
	int i;

	// init
	for (i=0; i < FSA_MAX_FSPERARCH; i++)
		dicoargv[i]=NULL;
	
	for (i=0; (i < FSA_MAX_FSPERARCH) && (cmdargv[i]); i++)
	{
		if (convert_string_to_dico(&tmpdico, &fsid, cmdargv[i], fscount)!=0)
		{	msgprintf(MSG_STACK, "convert_string_to_dico(%s) failed\n", cmdargv[i]);
			return -1;
		}
		else if ((fsid>=0) && (fsid < FSA_MAX_FSPERARCH)) // success
		{
			if (dicoargv[fsid]!=NULL)
			{	errprintf("you mentioned filesystem with id=%d multiple times, cannot continue\n", fsid);
				return -1;
			}
			else // ok
			{	
				dicoargv[fsid]=tmpdico;
			}
		}
		else
		{	errprintf("you must specify which filesystem to extract with 'id=x'\n");
			return -1;
		}
	}
	
	return 0;
}

int do_extract(char *archive, char *cmdargv[], int oper)
{
	struct s_dico *dicofsinfo[FSA_MAX_FSPERARCH];
	struct s_dico *dicoargv[FSA_MAX_FSPERARCH];
	pthread_t thread_decomp[FSA_MAX_COMPJOBS];
	char magic[FSA_SIZEOF_MAGIC+1];
	struct s_dico *dicomainhead=NULL;
	pthread_t thread_reader;
	struct s_archive ai;
	struct stat64 st;
	char *destdir;
	int errors=0;
	int ret=0;
	int i;
	
	// init archive
	archive_init(&ai);
	
	// init misc data struct to zero
	for (i=0; i<FSA_MAX_FSPERARCH; i++)
		dicoargv[i]=NULL;
	for (i=0; i<FSA_MAX_FSPERARCH; i++)
		dicofsinfo[i]=NULL;
	for (i=0; i<FSA_MAX_COMPJOBS; i++)
		thread_decomp[i]=0;
	thread_reader=0;
	
	// set archive path
	snprintf(ai.basepath, PATH_MAX, "%s", archive);
	
	// create decompression threads
	for (i=0; (i<g_options.compressjobs) && (i<FSA_MAX_COMPJOBS); i++)
	{
		if (pthread_create(&thread_decomp[i], NULL, thread_decomp_fct, (void*)&ai) != 0)
		{       errprintf("pthread_create(thread_decomp_fct) failed\n");
			ret=-1;
			goto do_extract_error;
		}
	}
	
	// create archive-reader thread
	if (pthread_create(&thread_reader, NULL, thread_reader_fct, (void*)&ai) != 0)
	{       errprintf("pthread_create(thread_reader_fct) failed\n");
		ret=-1;
		goto do_extract_error;
	}
	
	// read archive main header
	if (read_mainhead(&ai, &dicomainhead)<0)
	{       errprintf("read_mainhead(%s) failed\n", archive);
		ret=-1;
		goto do_extract_error;
	}
	
	if (oper==OPER_ARCHINFO && archinfo_show_mainhead(&ai, dicomainhead)!=0)
	{       errprintf("archinfo_show_mainhead(%s) failed\n", archive);
		ret=-1;
		goto do_extract_error;
	}
	
	// check that the operation requested on the command line matches the archive type
	switch (ai.archtype)
	{
		case ARCHTYPE_DIRECTORIES:
			if (oper==OPER_RESTFS)
			{	errprintf("this archive does not contain filesystems, cannot use \"restfs\". Try \"restdir\" instead.\n");
				ret=-1;
				goto do_extract_error;
			}
			break;
		case ARCHTYPE_FILESYSTEMS:
			if (oper==OPER_RESTDIR)
			{	errprintf("this archive does not contain simple directories, cannot use \"restdir\". Try \"restfs\" instead.\n");
				ret=-1;
				goto do_extract_error;
			}
			break;
		default:
			errprintf("this archive has an unknown type: %d, cannot continue\n", ai.archtype);
			ret=-1;
			goto do_extract_error;
	}
	
	// read the fsinfo header for each filesystem
	if (ai.archtype==ARCHTYPE_FILESYSTEMS)
	{
		// convert the command line arguments to dicos
		if (oper==OPER_RESTFS && convert_argv_to_dicos(dicoargv, cmdargv, ai.fscount)!=0)
		{	errprintf("convert_argv_to_dico() failed\n");
			ret=-1;
			goto do_extract_error;
		}
		
		for (i=0; (i < ai.fscount) && (i<FSA_MAX_FSPERARCH); i++)
		{
			// ---- read filesystem-header from archive
			if (queue_dequeue_header(&g_queue, &dicofsinfo[i], magic)<=0)
			{	errprintf("queue_dequeue_header() failed: cannot read filesystem-info dico\n");
				ret=-1;
				goto do_extract_error;
			}
			
			if (memcmp(magic, FSA_MAGIC_FSIN, FSA_SIZEOF_MAGIC)!=0)
			{	errprintf("header is not what we expected: found=[%s] and expected=[%s]\n", magic, FSA_MAGIC_FSIN);
				ret=-1;
				goto do_extract_error;
			}

			if (oper==OPER_ARCHINFO && archinfo_show_fshead(dicofsinfo[i], i)!=0)
			{	errprintf("archinfo_show_fshead() failed\n");
				ret=-1;
				goto do_extract_error;
			}
		}
	}
	
	if (oper==OPER_ARCHINFO)
	{
		set_stopfillqueue(); // don't need any more data
		while (queue_get_end_of_queue(&g_queue)==false) // let secondary threads exit
			queue_destroy_first_item(&g_queue);
		goto do_extract_success;
	}
	
	if (ai.archtype==ARCHTYPE_FILESYSTEMS)
	{
		// extract filesystem contents
		for (i=0; (i < ai.fscount) && (i < FSA_MAX_FSPERARCH) && (get_abort()==false); i++)
		{
			if (dicoargv[i]!=NULL) // that filesystem has been requested on the command line
			{
				msgprintf(1, "============= extracting filesystem %d =============\n", i);
				if (filesystem_extract(&ai, dicofsinfo[i], dicoargv[i], i)!=0)
				{	errprintf("extract_filesystem(%d) failed\n", i);
					goto do_extract_error;
				}
			}
			else // filesystem not specified on the command line
			{
				msgprintf(1, "============== skipping filesystem %d ==============\n\n", i);
				if (filesystem_skip(&ai)!=0)
				{	errprintf("filesystem_skip(%d) failed\n", i);
					goto do_extract_error;
				}
			}
		}
	}
	else if (ai.archtype==ARCHTYPE_DIRECTORIES)
	{
		destdir=cmdargv[0];
		if (stat64(destdir, &st)!=0)
		{	
			switch (errno)
			{
				case ENOENT:
					sysprintf("%s does not exist, cannot continue\n", destdir);
					break;
				default:
					sysprintf("fstat64(%d) failed\n", destdir);
					break;
			}
			goto do_extract_error;
		}
		
		if (!S_ISDIR(st.st_mode))
		{	errprintf("%s is not a valid directory, cannot continue\n", destdir);
			goto do_extract_error;
		}
		
		if (extract_read_objects(&ai, &errors, destdir, 0)!=0)
		{	errprintf("extract_read_objects(%s) failed\n", destdir);
			goto do_extract_error;
		}
	}
	else
	{	errprintf("unsupported archtype: %d\n", ai.archtype);
		goto do_extract_error;
	}
	
	if (get_abort()==false && get_stopfillqueue()==false)
		goto do_extract_success;
	
do_extract_error:
	msgprintf(2, "THREAD-MAIN2: exit error\n");
	set_stopfillqueue(); // don't need any more data
	while (queue_get_end_of_queue(&g_queue)==false) // let the secondary threads exit
		queue_destroy_first_item(&g_queue);
	msgprintf(2, "THREAD-MAIN2: queue is now empty\n");
	ret=-1;
	
do_extract_success:
	msgprintf(2, "THREAD-MAIN2: exit\n");
	
	for (i=0; (i<g_options.compressjobs) && (i<FSA_MAX_COMPJOBS); i++)
		if (thread_decomp[i] && pthread_join(thread_decomp[i], NULL) != 0)
			errprintf("pthread_join(thread_decomp) failed\n");
	
	if (thread_reader && pthread_join(thread_reader, NULL) != 0)
		errprintf("pthread_join(thread_reader) failed\n");
	
	for (i=0; i<FSA_MAX_FSPERARCH; i++)
		if (dicoargv[i]!=NULL)
			dico_destroy(dicoargv[i]);
	
	for (i=0; i<FSA_MAX_FSPERARCH; i++)
	{
		if (dicofsinfo[i]!=NULL)
		{
			dico_destroy(dicofsinfo[i]);
			dicofsinfo[i]=NULL;
		}
	}
	
	dico_destroy(dicomainhead);
	archive_destroy(&ai);
	return ret;
}
