/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "dico.h"
#include "common.h"
#include "fs_btrfs.h"
#include "uuid.h"

int btrfs_mkfs(struct s_dico *d, char *partition, char *filesystem)
{
	char command[2048];
	char buffer[2048];
	char options[2048];
	char temp[1024];
	u64 temp64;
	
	// there is no option that just displays the version and return 0 in mkfs.btrfs-0.16
	if (exec_command(command, sizeof(command), "which mkfs.btrfs")!=0)
	{	errprintf("mkfs.btrfs not found. please install btrfs-progs on your system or check the PATH.\n");
		return -1;
	}
	
	// ---- set the advanced filesystem settings from the dico
	memset(options, 0, sizeof(options));
	if (dico_get_string(d, 0, FSYSHEADKEY_FSLABEL, buffer, sizeof(buffer))==0 && strlen(buffer)>0)
	{	snprintf(temp, sizeof(temp), " -L %s ", buffer);
		bufcat(options, temp, sizeof(options));
		msgprintf(2, "FSYSHEADKEY_FSLABEL: mkfs_options+=[%s]\n", temp);
	}
	
	if (dico_get_u64(d, 0, FSYSHEADKEY_FSBTRFSSECTORSIZE, &temp64)==0)
	{	snprintf(temp, sizeof(temp), " -s %ld ", (long)temp64);
		bufcat(options, temp, sizeof(options));
		msgprintf(2, "FSYSHEADKEY_FSBTRFSSECTORSIZE: mkfs_options+=[%s]\n", temp);
	}
	
	if (exec_command(command, sizeof(command), "mkfs.btrfs %s %s", partition, options)!=0)
	{	errprintf("command [%s] failed\n", command);
		return -1;
	}
	
	return 0;
}

int btrfs_get_specific_features(struct s_dico *d, char *devname)
{
	struct btrfs_super_block sb;
	const char *str;
	u16 temp32;
	int ret=0;
	int fd;
	int res;
	
	if ((fd=open64(devname, O_RDONLY|O_LARGEFILE))<0)
	{	ret=-1;
		errprintf("cannot open(%s, O_RDONLY)\n", devname);
		goto btrfs_read_sb_return;
	}
	
	if (lseek(fd, BTRFS_SUPER_INFO_OFFSET, SEEK_SET)!=BTRFS_SUPER_INFO_OFFSET)
	{	ret=-2;
		errprintf("cannot lseek(fd, BTRFS_SUPER_INFO_OFFSET, SEEK_SET) on %s\n", devname);
		goto btrfs_read_sb_close;
	}
	
	res=read(fd, &sb, sizeof(sb));
	if (res!=sizeof(sb))
	{	ret=-3;
		errprintf("cannot read the btrfs superblock\n", devname);
		goto btrfs_read_sb_close;
	}
	
	if (strncmp((char*)&sb.magic, BTRFS_MAGIC, sizeof(sb.magic))!=0)
	{	ret=-4;
		errprintf("magic different from expectations superblock on %s: magic=[%.8s], expected=[%.8s]\n", devname, &sb.magic, BTRFS_MAGIC);
		goto btrfs_read_sb_close;
	}
	
	// ---- label
	msgprintf(3, "btrfs_label=[%s]\n", sb.label);
	dico_add_string(d, 0, FSYSHEADKEY_FSLABEL, (char*)sb.label);
	
	// ---- uuid
	if ((str=e2p_uuid2str(sb.dev_item.fsid))!=NULL)
		dico_add_string(d, 0, FSYSHEADKEY_FSUUID, str);
	msgprintf(3, "btrfs_uuid=[%s]\n", str);
	
	// ---- sector size
	temp32=le32_to_cpu(sb.sectorsize);
	dico_add_u64(d, 0, FSYSHEADKEY_FSBTRFSSECTORSIZE, temp32);
	msgprintf(3, "btrfs_sectorsize=[%ld]\n", (long)temp32);
	
btrfs_read_sb_close:
	close(fd);
btrfs_read_sb_return:
	return ret;
}
