/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

// Parts of this file comes from e2fsprogs-1.41.3

#include "fsarchiver.h"

#include <stdio.h>
#include <ext2fs/ext2_fs.h>
#include <ext2fs/ext2fs.h>

#include "dico.h"
#include "common.h"
#include "uuid.h"
#include "fs_ext2.h"

struct mntopt
{       unsigned int    mask;
        const char      *string;
};

struct feature 
{
        int             compat;
        unsigned int    mask;
        const char      *string;
};

static struct mntopt mntopt_list[] = {
	{ EXT2_DEFM_DEBUG,	"debug" },
	{ EXT2_DEFM_BSDGROUPS,	"bsdgroups" },
	{ EXT2_DEFM_XATTR_USER,	"user_xattr" },
	{ EXT2_DEFM_ACL,	"acl" },
	{ EXT2_DEFM_UID16,	"uid16" },
	{ EXT3_DEFM_JMODE_DATA, "journal_data" },
	{ EXT3_DEFM_JMODE_ORDERED, "journal_data_ordered" },
	{ EXT3_DEFM_JMODE_WBACK, "journal_data_writeback" },
	{ 0, 0 },
};

static struct feature feature_list[] = {
        {       E2P_FEATURE_COMPAT, EXT2_FEATURE_COMPAT_DIR_PREALLOC,
                        "dir_prealloc" },
        {       E2P_FEATURE_COMPAT, EXT3_FEATURE_COMPAT_HAS_JOURNAL,
                        "has_journal" },
        {       E2P_FEATURE_COMPAT, EXT2_FEATURE_COMPAT_IMAGIC_INODES,
                        "imagic_inodes" },
        {       E2P_FEATURE_COMPAT, EXT2_FEATURE_COMPAT_EXT_ATTR,
                        "ext_attr" },
        {       E2P_FEATURE_COMPAT, EXT2_FEATURE_COMPAT_DIR_INDEX,
                        "dir_index" },
        {       E2P_FEATURE_COMPAT, EXT2_FEATURE_COMPAT_RESIZE_INODE,
                        "resize_inode" },
        {       E2P_FEATURE_COMPAT, EXT2_FEATURE_COMPAT_LAZY_BG,
                        "lazy_bg" },

        {       E2P_FEATURE_RO_INCOMPAT, EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER,
                        "sparse_super" },
        {       E2P_FEATURE_RO_INCOMPAT, EXT2_FEATURE_RO_COMPAT_LARGE_FILE,
                        "large_file" },
        {       E2P_FEATURE_RO_INCOMPAT, EXT4_FEATURE_RO_COMPAT_HUGE_FILE,
                        "huge_file" },
        {       E2P_FEATURE_RO_INCOMPAT, EXT4_FEATURE_RO_COMPAT_GDT_CSUM,
                        "uninit_bg" },
        {       E2P_FEATURE_RO_INCOMPAT, EXT4_FEATURE_RO_COMPAT_GDT_CSUM,
                        "uninit_groups" },
        {       E2P_FEATURE_RO_INCOMPAT, EXT4_FEATURE_RO_COMPAT_DIR_NLINK,
                        "dir_nlink" },
        {       E2P_FEATURE_RO_INCOMPAT, EXT4_FEATURE_RO_COMPAT_EXTRA_ISIZE,
                        "extra_isize" },

        {       E2P_FEATURE_INCOMPAT, EXT2_FEATURE_INCOMPAT_COMPRESSION,
                        "compression" },
        {       E2P_FEATURE_INCOMPAT, EXT2_FEATURE_INCOMPAT_FILETYPE,
                        "filetype" },
        {       E2P_FEATURE_INCOMPAT, EXT3_FEATURE_INCOMPAT_RECOVER,
                        "needs_recovery" },
        {       E2P_FEATURE_INCOMPAT, EXT3_FEATURE_INCOMPAT_JOURNAL_DEV,
                        "journal_dev" },
        {       E2P_FEATURE_INCOMPAT, EXT3_FEATURE_INCOMPAT_EXTENTS,
                        "extent" },
        {       E2P_FEATURE_INCOMPAT, EXT3_FEATURE_INCOMPAT_EXTENTS,
                        "extents" },
        {       E2P_FEATURE_INCOMPAT, EXT2_FEATURE_INCOMPAT_META_BG,
                        "meta_bg" },
        {       E2P_FEATURE_INCOMPAT, EXT4_FEATURE_INCOMPAT_64BIT,
                        "64bit" },
        {       E2P_FEATURE_INCOMPAT, EXT4_FEATURE_INCOMPAT_FLEX_BG,
                        "flex_bg"},
        {       0, 0, 0 },
};

// all the ext{2/3/4} filesystem features that can be either enabled or disabled when we run mkfs
static char *mkfs_features[] = {"large_file", "dir_index", "filetype", "flex_bg", "has_journal",
	"journal_dev", "extent", "uninit_bg", "resize_inode", "sparse_super", NULL};

const char *e2p_mntopt2string(unsigned int mask)
{
	struct mntopt *f;
	static char buf[20];
	int fnum;

	for (f = mntopt_list; f->string; f++) 
	{
		if (mask == f->mask)
			return f->string;
	}
	for (fnum = 0; mask >>= 1; fnum++);
	sprintf(buf, "MNTOPT_%d", fnum);
	return buf;
}

const char *e2p_feature2string(int compat, unsigned int mask)
{
        struct feature  *f;
        static char buf[20];
        char    fchar;
        int     fnum;

        for (f = feature_list; f->string; f++) {
                if ((compat == f->compat) &&
                    (mask == f->mask))
                        return f->string;
        }
        switch (compat) {
        case  E2P_FEATURE_COMPAT:
                fchar = 'C';
                break;
        case E2P_FEATURE_INCOMPAT:
                fchar = 'I';
                break;
        case E2P_FEATURE_RO_INCOMPAT:
                fchar = 'R';
                break;
        default:
                fchar = '?';
                break;
        }
        for (fnum = 0; mask >>= 1; fnum++);
        sprintf(buf, "FEATURE_%c%d", fchar, fnum);
        return buf;
}

int ext2_mkfs(struct s_dico *d, char *partition, char *filesystem)
{
	char buffer[2048];
	char command[2048];
	char options[2048];
	char temp[1024];
	char progname[64];
	u64 temp64;
	int i;
	
	// ---- choose the right mkfs program
	if (strcmp(filesystem, "ext2")==0)
		snprintf(progname, sizeof(progname), "mkfs.ext2");
	else if (strcmp(filesystem, "ext3")==0)
		snprintf(progname, sizeof(progname), "mkfs.ext3");
	else if (strcmp(filesystem, "ext4")==0)
		snprintf(progname, sizeof(progname), "mkfs.ext4");
	else
		return -1;
	
	// ---- check that mkfs is installed
	if (exec_command(command, sizeof(command), "%s -V", progname)!=0)
	{	errprintf("%s not found. please install a recent e2fsprogs on your system or check the PATH.\n", progname);
		return -1;
	}
	
	// ---- forces the creation of a journal on ext3 and ext4
	if ((strcmp(filesystem, "ext3")==0) || (strcmp(filesystem, "ext4")==0))
	{	snprintf(temp, sizeof(temp), " -j ");
		bufcat(options, temp, sizeof(options));
		msgprintf(3, "mkfs_options+=[%s]\n", temp);
	}
	
	// ---- set the advanced filesystem settings from the dico
	memset(options, 0, sizeof(options));
	if (dico_get_string(d, 0, FSYSHEADKEY_FSLABEL, buffer, sizeof(buffer))==0 && strlen(buffer)>0)
	{	snprintf(temp, sizeof(temp), " -L %s ", buffer);
		bufcat(options, temp, sizeof(options));
		msgprintf(2, "FSYSHEADKEY_FSLABEL: mkfs_options+=[%s]\n", temp);
	}
	
	if (dico_get_u64(d, 0, FSYSHEADKEY_FSEXTBLOCKSIZE, &temp64)==0)
	{
		snprintf(temp, sizeof(temp), " -b %ld ", (long)temp64);
		bufcat(options, temp, sizeof(options));
		msgprintf(2, "FSYSHEADKEY_FSEXTBLOCKSIZE: mkfs_options+=[%s]\n", temp);
	}
	
	if (dico_get_u64(d, 0, FSYSHEADKEY_FSINODESIZE, &temp64)==0)
	{
		snprintf(temp, sizeof(temp), " -I %ld ", (long)temp64);
		bufcat(options, temp, sizeof(options));
		msgprintf(2, "FSYSHEADKEY_FSINODESIZE: mkfs_options+=[%s]\n", temp);
	}
	
	if (dico_get_string(d, 0, FSYSHEADKEY_FSEXTFEATURES, buffer, sizeof(buffer))==0 && strlen(buffer)>0)
	{	
		// disable all features which are not in the list (disable new default features that may break old systems)
		msgprintf(3, "Original filesystem features: [%s]\n----\n", buffer);
		for (i=0; mkfs_features[i]; i++)
		{
			if (strstr(buffer, mkfs_features[i])==NULL)
			{	snprintf(temp, sizeof(temp), ",^%s", mkfs_features[i]);
				bufcat(buffer, temp, sizeof(buffer));
				msgprintf(3, "--> feature [%s]=NO --> buf=[%s]\n", mkfs_features[i], buffer);
			}
			else
			{	msgprintf(3, "--> feature [%s]=YES\n", mkfs_features[i]);
			}
		}
		
		snprintf(temp, sizeof(temp), " -O %s ", buffer);
		bufcat(options, temp, sizeof(options));
		msgprintf(2, "FSYSHEADKEY_FSEXTFEATURES: mkfs_options+=[%s]\n", temp);
	}
	
	msgprintf(2, "exec: %s\n", command);
	if (exec_command(command, sizeof(command), "%s %s %s", progname, partition, options)!=0)
	{	errprintf("command [%s] failed\n", command);
		return -1;
	}
	
	// ---- use tune2fs to set the other advanced options
	memset(options, 0, sizeof(options));
	if (dico_get_string(d, 0, FSYSHEADKEY_FSUUID, buffer, sizeof(buffer))==0 && strlen(buffer)==36)
	{	snprintf(temp, sizeof(temp), " -U %s ", buffer);
		bufcat(options, temp, sizeof(options));
		msgprintf(2, "FSYSHEADKEY_FSUUID: tune2fs_options+=[%s]\n", temp);
	}
	
	if (dico_get_string(d, 0, FSYSHEADKEY_FSEXTDEFMNTOPT, buffer, sizeof(buffer))==0 && strlen(buffer)>0)
	{	snprintf(temp, sizeof(temp), " -o %s ", buffer);
		bufcat(options, temp, sizeof(options));
		msgprintf(2, "FSYSHEADKEY_FSEXTDEFMNTOPT: tune2fs_options+=[%s]\n", temp);
	}
	
	if (options[0])
	{
		if (exec_command(command, sizeof(command), "tune2fs %s %s", partition, options)!=0)
		{	errprintf("command [%s] failed\n", command);
			return -1;
		}
		
		// TODO: remove this fsck when the bug if fixed in the official e2fsprogs release
		// run mkfs to workaround an tune2fs bug in e2fsprogs-1.41.3 on ext4
		// http://article.gmane.org/gmane.comp.file-systems.ext4/11181
		if (strcmp(filesystem, "ext4")==0)
		{
			if (exec_command(command, sizeof(command), "fsck.ext4 -fy %s", partition)!=0)
			{	errprintf("command [%s] failed\n", command);
				return -1;
			}
		}
	}
	
	return 0;
}

int ext2_get_specific_features(struct s_dico *d, char *devname)
{
	blk_t use_superblock=0;
	int use_blocksize=0;
	const char *str;
	ext2_filsys fs;
	u32 mask, m;
	u32 *mask2;
	char temp[1024];
	char mntopt[1024];
	char features[1024];
	char label[80];
	int count;
	int flags;
	int i, j;
	
	// ---- open partition
	flags = EXT2_FLAG_JOURNAL_DEV_OK | EXT2_FLAG_SOFTSUPP_FEATURES;
	if (ext2fs_open(devname, flags, use_superblock,  use_blocksize, unix_io_manager, &fs)!=0)
	{	errprintf("ext2fs_open() failed\n");
		return -1;
	}
	
	// --- label
	memset(label, 0, sizeof(label));
	if (fs->super->s_volume_name[0])
	{	memset(label, 0, sizeof(label));
		strncpy(label, fs->super->s_volume_name, sizeof(fs->super->s_volume_name));
	}
	dico_add_string(d, 0, FSYSHEADKEY_FSLABEL, label);
	
	// ---- uuid
	if ((str=e2p_uuid2str(fs->super->s_uuid))!=NULL)
		dico_add_string(d, 0, FSYSHEADKEY_FSUUID, str);
	
	// ---- block size
	dico_add_u64(d, 0, FSYSHEADKEY_FSEXTBLOCKSIZE, EXT2_BLOCK_SIZE(fs->super));
	
	// ---- inode size
	if (fs->super->s_rev_level >= EXT2_DYNAMIC_REV)
		dico_add_u64(d, 0, FSYSHEADKEY_FSINODESIZE, fs->super->s_inode_size);
	
	// ---- default mount options
	memset(mntopt, 0, sizeof(mntopt));
	count=0;
	mask=fs->super->s_default_mount_opts;
	if (mask & EXT3_DEFM_JMODE)
	{	
		snprintf(temp, sizeof(temp), "%s", e2p_mntopt2string(mask & EXT3_DEFM_JMODE));
		bufcat(mntopt, temp, sizeof(mntopt));
		count++;
	}
	for (i=0, m=1; i < 32; i++, m<<=1)
	{
		if (m & EXT3_DEFM_JMODE)
			continue;
		if (mask & m)
		{
			if (count++) bufcat(mntopt, ",", sizeof(mntopt));
			snprintf(temp, sizeof(temp), "%s", e2p_mntopt2string(m));
			bufcat(mntopt, temp, sizeof(mntopt));
		}
	}
	dico_add_string(d, 0, FSYSHEADKEY_FSEXTDEFMNTOPT, mntopt);
	msgprintf(2, "default mount options: [%s]\n", mntopt);
	
	// ---- filesystem features
	memset(features, 0, sizeof(features));
        mask2 = &fs->super->s_feature_compat;
        for (count=0, i=0; i <3; i++, mask2++)
	{
                for (j=0, m=1; j < 32; j++, m<<=1)
		{
                        if (*mask2 & m)
			{
				snprintf(temp, sizeof(temp), "%s", e2p_feature2string(i, m));
				if (strcmp(temp, "needs_recovery")!=0) // ignore 'needs_recovery' that mkfs cannot use
				{
					if (count++) bufcat(features, ",", sizeof(features));
					bufcat(features, temp, sizeof(features));
				}
                        }
                }
        }
	dico_add_string(d, 0, FSYSHEADKEY_FSEXTFEATURES, features);
	msgprintf(2, "filesystem features: [%s]\n", features);
	
	// ---- close partition
	ext2fs_close(fs);
	
	return 0;
}
