/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "dico.h"
#include "common.h"
#include "fs_jfs.h"
#include "uuid.h"

int jfs_mkfs(struct s_dico *d, char *partition, char *filesystem)
{
	char command[2048];
	char buffer[2048];
	char options[2048];
	char temp[1024];
	
	// ---- check mkfs.reiser4 is available
	if (exec_command(command, sizeof(command), "mkfs.jfs -V")!=0)
	{	errprintf("mkfs.jfs not found. please install jfsutils on your system or check the PATH.\n");
		return -1;
	}
	
	// ---- set the advanced filesystem settings from the dico
	memset(options, 0, sizeof(options));
	if (dico_get_string(d, 0, FSYSHEADKEY_FSLABEL, buffer, sizeof(buffer))==0 && strlen(buffer)>0)
	{	snprintf(temp, sizeof(temp), " -L %s ", buffer);
		bufcat(options, temp, sizeof(options));
		msgprintf(2, "FSYSHEADKEY_FSLABEL: mkfs_options+=[%s]\n", temp);
	}
	
	if (exec_command(command, sizeof(command), "mkfs.jfs -q %s %s", options, partition)!=0)
	{	errprintf("command [%s] failed\n", command);
		return -1;
	}
	
	// ---- use jfs_tune to set the other advanced options
	memset(options, 0, sizeof(options));
	if (dico_get_string(d, 0, FSYSHEADKEY_FSUUID, buffer, sizeof(buffer))==0 && strlen(buffer)==36)
	{	snprintf(temp, sizeof(temp), " -U %s ", buffer);
		bufcat(options, temp, sizeof(options));
		msgprintf(2, "FSYSHEADKEY_FSUUID: jfs_tune_options+=[%s]\n", temp);
	}
	
	if (options[0])
	{
		if (exec_command(command, sizeof(command), "jfs_tune %s %s", partition, options)!=0)
		{	errprintf("command [%s] failed\n", command);
			return -1;
		}
	}
	
	return 0;
}

int jfs_get_specific_features(struct s_dico *d, char *devname)
{
	struct jfs_superblock sb;
	const char *str;
	int ret=0;
	int fd;
	int res;
	
	fd=open64(devname, O_RDONLY|O_LARGEFILE);
	if (fd<0)
	{	ret=-1;
		errprintf("cannot open(%s, O_RDONLY)\n", devname);
		goto jfs_get_specific_features_return;
	}
	
	if (lseek(fd, JFS_SUPER1_OFF, SEEK_SET)!=JFS_SUPER1_OFF)
	{	ret=-2;
		errprintf("cannot lseek(fd, JFS_SUPER1_OFF, SEEK_SET) on %s\n", devname);
		goto jfs_get_specific_features_close;
	}
	
	res=read(fd, &sb, sizeof(sb));
	if (res!=sizeof(sb))
	{	ret=-3;
		errprintf("cannot read the jfs superblock\n", devname);
		goto jfs_get_specific_features_close;
	}
	
	if (strncmp(sb.s_magic, JFS_MAGIC, strlen(JFS_MAGIC)) != 0)
	{	ret=-4;
		errprintf("magic different from expectations superblock on %s: magic=[%s], expected=[%s]\n", devname, sb.s_magic, JFS_MAGIC);
		goto jfs_get_specific_features_close;
	}
	msgprintf(3, "jfs_magic=[%s]\n", sb.s_magic);
	
	// ---- label
	msgprintf(3, "jfs_label=[%s]\n", sb.s_label);
	dico_add_string(d, 0, FSYSHEADKEY_FSLABEL, (char*)sb.s_label);
	
	// ---- uuid
	if ((str=e2p_uuid2str(sb.s_uuid))!=NULL)
		dico_add_string(d, 0, FSYSHEADKEY_FSUUID, str);
	msgprintf(3, "jfs_uuid=[%s]\n", str);
		
jfs_get_specific_features_close:
	close(fd);
jfs_get_specific_features_return:
	return ret;
}

