/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "dico.h"
#include "common.h"
#include "fs_xfs.h"
#include "uuid.h"

int xfs_mkfs(struct s_dico *d, char *partition, char *filesystem)
{
	char command[2048];
	char buffer[2048];
	char options[2048];
	char temp[1024];
	u64 temp64;

	if (exec_command(command, sizeof(command), "mkfs.xfs -V")!=0)
	{	errprintf("mkfs.xfs not found. please install xfsprogs on your system or check the PATH.\n");
		return -1;
	}

	memset(options, 0, sizeof(options));
	if (dico_get_string(d, 0, FSYSHEADKEY_FSLABEL, buffer, sizeof(buffer))==0 && strlen(buffer)>0)
	{	snprintf(temp, sizeof(temp), " -L %s ", buffer);
		bufcat(options, temp, sizeof(options));
		msgprintf(2, "FSYSHEADKEY_FSLABEL: mkfs_options+=[%s]\n", temp);
	}
	
	if ((dico_get_u64(d, 0, FSYSHEADKEY_FSXFSBLOCKSIZE, &temp64)==0) && (temp64%512==0) && (temp64>=512) && (temp64<=65536))
	{
		snprintf(temp, sizeof(temp), " -b size=%ld ", (long)temp64);
		bufcat(options, temp, sizeof(options));
		msgprintf(2, "FSYSHEADKEY_FSXFSBLOCKSIZE: mkfs_options+=[%s]\n", temp);
	}

	if (exec_command(command, sizeof(command), "mkfs.xfs -f %s %s", partition, options)!=0)
	{	errprintf("command [%s] failed\n", command);
		return -1;
	}

	// ---- use xfs_admin to set the other advanced options
	memset(options, 0, sizeof(options));
	if (dico_get_string(d, 0, FSYSHEADKEY_FSUUID, buffer, sizeof(buffer))==0 && strlen(buffer)==36)
	{	snprintf(temp, sizeof(temp), " -U %s ", buffer);
		bufcat(options, temp, sizeof(options));
		msgprintf(2, "FSYSHEADKEY_FSUUID: xfs_admin_options+=[%s]\n", temp);
	}
	
	if (options[0])
	{
		if (exec_command(command, sizeof(command), "xfs_admin %s %s", options, partition)!=0)
		{	errprintf("command [%s] failed\n", command);
			return -1;
		}
	}

	return 0;	
}

int xfs_get_specific_features(struct s_dico *d, char *devname)
{
	struct xfs_sb sb;
	const char *str;
	u32 temp32;
	int ret=0;
	int fd;
	int res;

	fd=open64(devname, O_RDONLY|O_LARGEFILE);
	if (fd<0)
	{	ret=-1;
		goto xfs_read_sb_return;
	}

	res=read(fd, &sb, sizeof(sb));
	if (res!=sizeof(sb))
	{	ret=-1;
		goto xfs_read_sb_close;
	}

	// ---- check it's an XFS file system
	if (be32_to_cpu(sb.sb_magicnum) != XFS_SUPER_MAGIC)
	{	ret=-1;
		msgprintf(3, "sb.sb_magicnum!=XFS_SUPER_MAGIC\n");
		goto xfs_read_sb_close;
	}
	
	// ---- label
	msgprintf(3, "xfs_label=[%s]\n", sb.sb_fname);
	dico_add_string(d, 0, FSYSHEADKEY_FSLABEL, (char*)sb.sb_fname);
	
	// ---- uuid
	if ((str=e2p_uuid2str(&sb.sb_uuid))!=NULL)
		dico_add_string(d, 0, FSYSHEADKEY_FSUUID, str);
	msgprintf(3, "xfs_uuid=[%s]\n", str);

	// ---- block size
	temp32=be32_to_cpu(sb.sb_blocksize);
	if ((temp32%512!=0) || (temp32<512) || (temp32>65536))
	{	ret=-1;
		msgprintf(3, "xfs_blksize=[%ld] is an invalid xfs block size\n", (long)temp32);
		goto xfs_read_sb_close;
	}
	dico_add_u64(d, 0, FSYSHEADKEY_FSXFSBLOCKSIZE, temp32);
	msgprintf(3, "xfs_blksize=[%ld]\n", (long)temp32);
	
xfs_read_sb_close:
	close(fd);
xfs_read_sb_return:
	return ret;
}
