/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <endian.h>
#include <byteswap.h>
#include <signal.h>
#include <linux/limits.h>

#include "dico.h"
#include "common.h"
#include "extract.h"
#include "create.h"
#include "archinfo.h"
#include "syncthread.h"
#include "comp_lzo.h"

struct s_options g_options;
int g_logfile=-1;

char *valid_magic[]={FSA_MAGIC_MAIN, FSA_MAGIC_VOLH, FSA_MAGIC_VOLF, FSA_MAGIC_FSIN, 
	FSA_MAGIC_FSYB, FSA_MAGIC_DATF, FSA_MAGIC_OBJT, FSA_MAGIC_BLKH, FSA_MAGIC_FILF, NULL};

void usage(char *progname)
{
	char optlzo[32];
	char optlzma[32];
	
	// init
	memset(optlzo, 0, sizeof(optlzo));
	memset(optlzma, 0, sizeof(optlzma));
	
#ifdef OPTION_LZO_SUPPORT
	snprintf(optlzo, sizeof(optlzo), "1 (lzo), ");
#endif // OPTION_LZO_SUPPORT
	
#ifdef OPTION_LZMA_SUPPORT
	snprintf(optlzma, sizeof(optlzma), ", 4 (lzma)");
#endif // OPTION_LZMA_SUPPORT
	
	msgprintf(0, "====> fsarchiver version %s (%s) - http://www.fsarchiver.org <====\n", FSA_VERSION, FSA_RELDATE);
	msgprintf(0, "Distributed under the GPL v2 license (GNU General Public License v2).\n");
	msgprintf(0, "<usage>\n");
	msgprintf(0, " * usage: %s [<options>] <command> <archive> [<part1> [<part2> [...]]]\n", progname);
	msgprintf(0, "<commands>\n");
	msgprintf(0, " * savefs: save filesystems to an archive file (backup a partition to a file)\n");
	msgprintf(0, " * restfs: restore filesystems from an archive (overwrites the existing data)\n");
	msgprintf(0, " * savedir: save directories to the archive (similar to a compressed tarball)\n");
	msgprintf(0, " * restdir: restore data from an archive which is not based on a filesystem\n");
	msgprintf(0, " * archinfo: show information about an existing archive file and its contents\n");
	msgprintf(0, "<examples>\n");
	msgprintf(0, " * save only one filesystem (/dev/sda1) to an archive:\n");
	msgprintf(0, "   fsarchiver savefs /data/myarchive1.fsa /dev/sda1\n");
	msgprintf(0, " * save two filesystems (/dev/sda1 and /dev/sdb1) to an archive:\n");
	msgprintf(0, "   fsarchiver savefs /data/myarchive2.fsa /dev/sda1 /dev/sdb1\n");
	msgprintf(0, " * restore the first filesystem from an archive (first = number 0):\n");
	msgprintf(0, "   fsarchiver restfs /data/myarchive2.fsa id=0,dest=/dev/sda1\n");
	msgprintf(0, " * restore the second filesystem from an archive (second = number 1):\n");
	msgprintf(0, "   fsarchiver restfs /data/myarchive2.fsa id=1,dest=/dev/sdb1\n");
	msgprintf(0, " * restore two filesystems from an archive (number 0 and 1):\n");
	msgprintf(0, "   fsarchiver restfs /data/arch2.fsa id=0,dest=/dev/sda1 id=1,dest=/dev/sdb1\n");
	msgprintf(0, " * restore a filesystem from an archive and convert it to reiserfs:\n");
	msgprintf(0, "   fsarchiver restfs /data/myarchive1.fsa id=0,dest=/dev/sda1,mkfs=reiserfs\n");
	msgprintf(0, " * save the contents of /usr/src/linux to an archive (similar to tar):\n");
	msgprintf(0, "   fsarchiver savedir /data/linux-sources.fsa /usr/src/linux\n");
	msgprintf(0, " * save a filesystem (/dev/sda1) to an archive splitted into volumes of 680MB:\n");
	msgprintf(0, "   fsarchiver savefs -s 680 /data/myarchive1.fsa /dev/sda1\n");
	msgprintf(0, " * extract an archive made of simple files to /tmp/extract:\n");
	msgprintf(0, "   fsarchiver restdir /data/linux-sources.fsa /tmp/extract\n");
	msgprintf(0, " * show information about an archive and its file systems:\n");
	msgprintf(0, "   fsarchiver archinfo /data/myarchive2.fsa\n");
	msgprintf(0, "<options>\n");
	msgprintf(0, " -o: overwrite the archive if it already exists\n");
	msgprintf(0, " -v: verbose mode (can be used several times to increase the level of details)\n");
	msgprintf(0, " -d: debug mode (can be used several times to increase the level of details)\n");
	msgprintf(0, " -A: allow to save a filesystem which is mounted in read-write\n");
	msgprintf(0, " -z <level>: compression levels: %s2 (gzip), 3 (bzip2)%s\n", optlzo, optlzma);
	msgprintf(0, " -s <mbsize>: split the archive into several files of <mbsize> megabytes each\n");
	msgprintf(0, " -j <count>: create more than one compression thread. useful on multi-core cpu\n");
	msgprintf(0, " -h: show help and information about how to use fsarchiver\n");
	msgprintf(0, " -V: show program version and exit\n");
	msgprintf(0, "Warnings:\n");
	msgprintf(0, " * this a development version, don't use for critical data.\n");
	msgprintf(0, " * the archive format will change in the future.\n");
}

int main(int argc, char **argv)
{
	sigset_t mask_set;
	char *command=NULL;
	char *archive=NULL;
	char *partition[32];
	char *progname;
	int ret=0;
	int argcok;
	int cmd;
	int i;
	int c;
	
	// init
	progname=argv[0];
	memset(partition, 0, sizeof(partition));
	memset(&g_options, 0, sizeof(struct s_options));
	
	// set default options
	g_options.overwrite=0;
	g_options.allowsaverw=0;
	g_options.verboselevel=0;
	g_options.debuglevel=0;
	g_options.compressjobs=1;
	g_options.compresslevel=COMPRESS_GZIP;

	if (geteuid()!=0)
	{	errprintf("%s must be run as root. cannot continue.\n", progname);
		return 1;
	}
	
	while ((c = getopt(argc, argv, "oAvdz:j:hVs:")) != -1)
	{
		switch (c)
		{
			case 'o': // overwrite existing archives
				g_options.overwrite=1;
				break;
			case 'A': // allows to backup read/write mounted partition
				g_options.allowsaverw=1;
				break;
			case 'v': // verbose mode
				g_options.verboselevel++;
				break;
			case 'V': // version
				msgprintf(0, "fsarchiver %s (%s)\n", FSA_VERSION, FSA_RELDATE);
				return 0;
			case 'd': // debug mode
				g_options.debuglevel++;
				break;
			case 'j': // compression jobs
				g_options.compressjobs=atoi(optarg);
				if (g_options.compressjobs<1 || g_options.compressjobs>FSA_MAX_COMPJOBS)
				{
					errprintf("[%s] is not a valid job number. Must be between 1 and %d\n", optarg, FSA_MAX_COMPJOBS);
					usage(progname);
					return 1;
				}
				break;
			case 's': // split archive into several volumes
				g_options.splitsize=atol(optarg)*1024*1024;
				if (g_options.splitsize==0)
				{
					errprintf("argument of option -s is invalid (%s). It must be a valid integer\n", optarg);
					usage(progname);
					return 1;
				}
				break;
			case 'z': // compression level
				switch (atoi(optarg))
				{
#ifdef OPTION_LZO_SUPPORT
					case 1:
						g_options.compresslevel=COMPRESS_LZO;
						break;
#endif // OPTION_LZO_SUPPORT
					case 2:
						g_options.compresslevel=COMPRESS_GZIP;
						break;
					case 3:
						g_options.compresslevel=COMPRESS_BZIP2;
						break;
#ifdef OPTION_LZMA_SUPPORT
					case 4:
						g_options.compresslevel=COMPRESS_LZMA;
						break;
#endif
					default:
						errprintf("[%s] is not a valid compression level\n", optarg);
						usage(progname);
						return 1;
				};
				break;
			case 'h': // help
			default:
				usage(progname);
				return 1;
		}
	}
	
	argc -= optind;
	argv += optind;
	
	// in all cases we need at least 2 parameters
	if (argc < 2)
	{	fprintf(stderr, "there are not enough arguments on the command line\n");
		usage(progname);
		return 1;
	}
	else // mandatory and unique parameters
	{	command=*argv++, argc--;
		archive=*argv++, argc--;
	}
	
	// convert commands to integers
	if (strcmp(command, "savefs")==0)
	{	cmd=OPER_SAVEFS;
		argcok=(argc>0);
	}
	else if (strcmp(command, "restfs")==0)
	{	cmd=OPER_RESTFS;
		argcok=(argc>0);
	}
	else if (strcmp(command, "savedir")==0)
	{	cmd=OPER_SAVEDIR;
		argcok=(argc>0);
	}
	else if (strcmp(command, "restdir")==0)
	{	cmd=OPER_RESTDIR;
		argcok=(argc==1);
	}
	else if (strcmp(command, "archinfo")==0)
	{	cmd=OPER_ARCHINFO;
		argcok=(argc==0);
	}
	else // command not found
	{	errprintf("[%s] is not a valid command.\n", command);
		usage(progname);
		return 1;
	}
	
	// check there are enough parameters on the cmd line
	if (argcok!=true)
	{	errprintf("invalid arguments on the command line\n");
		usage(progname);
		return 1;
	}
	
	// open debug logfile if requested
	if (g_options.debuglevel>0)
	{	mkdir_recursive("/var/log");
		msgprintf(1, "g_options.debuglevel=%d\n", g_options.debuglevel);
		g_logfile=open64("/var/log/fsarchiver.log", O_RDWR|O_CREAT|O_TRUNC|O_LARGEFILE, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
		if (g_logfile>=0)
		{ 
			msgprintf(1, "Creating logfile in /var/log/fsarchiver.log\n");
			msgprintf(1, "Running fsarchiver version=[%s], fileformat=[%s]\n", FSA_VERSION, FSA_FILEFORMAT);
		}
		else
		{
			sysprintf("Cannot create logfile in /var/log/fsarchiver.log\n");
		}
	}
	
	// init the queue
	queue_init(&g_queue, FSA_MAX_QUEUESIZE);
	
#ifdef OPTION_LZO_SUPPORT
	// init the lzo library
	if (lzo_init() != LZO_E_OK)
	{	errprintf("internal error - lzo_init() failed\n");
		return 1;
	}
#endif // OPTION_LZO_SUPPORT
	
	// list of partitions to backup/restore
	for (i=0; (i < argc) && (argv[i]); i++)
		partition[i]=argv[i];
	
	// install signal handlers
	sigemptyset(&mask_set);
	sigaddset(&mask_set, SIGINT);
	sigaddset(&mask_set, SIGTERM);
	sigprocmask(SIG_SETMASK, &mask_set, NULL);
	
	switch (cmd)
	{
		case OPER_SAVEFS:
			ret=do_create(archive, partition, ARCHTYPE_FILESYSTEMS);
			break;
		case OPER_SAVEDIR:
			ret=do_create(archive, partition, ARCHTYPE_DIRECTORIES);
			break;
		case OPER_RESTFS:
		case OPER_RESTDIR:
		case OPER_ARCHINFO:
			ret=do_extract(archive, partition, cmd);
			break;
		default:
			errprintf("[%s] is not a valid command.\n", command);
			usage(progname);
			break;
			ret=1;
	};
	
	// destroy the queue
	queue_destroy(&g_queue);
	
	// close the debug logfile if it has been open
	if (g_logfile>=0)
		close(g_logfile);
	
	return !ret;
}
