/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/vfs.h>
#include <sys/utsname.h>
#include <linux/limits.h>
#include <sys/mount.h>

#include "common.h"
#include "mount.h"

int get_filesystem_info(char *dev, char *mnt, char *text, int textlen)
{
	struct statfs stf;
	struct stat64 st;
	
	// init
	memset(text, 0, textlen);

	if (stat64(dev, &st)!=0)
	{	//sysprintf("Cannot stat %s\n", dev);
		return -1;
	}

	if (statfs(dev, &stf)!=0)
	{	//sysprintf("Cannot statfs %s\n", dev);
		return -1;
	}
	
	if (!S_ISBLK(st.st_mode))
	{
		//printf("%s is not a block device\n", dev);
		return -1;
	}
	else
	{	//printf("%s is a block device\n", dev);
		return 0;
	}
}

int get_filesystem_rw_status(char *options)
{
	char temp[FSA_MAX_FSNAMELEN];
	char delims[]=",";
	char *saveptr;
	char *result;

	snprintf(temp, sizeof(temp), "%s", options);

	result=strtok_r(temp, delims, &saveptr);
	while(result != NULL)
	{	if (strcmp(result, "rw")==0)
			return 1; // true
		result = strtok_r(NULL, delims, &saveptr);
	}
	
	return 0; // false
}

int devcmp(char *dev1, char *dev2)
{
	struct stat64 st1, st2;
	
	if (stat64(dev1, &st1)!=0 || stat64(dev2, &st2)!=0)
		return -1;

	if (!S_ISBLK(st1.st_mode) || !S_ISBLK(st2.st_mode))
		return -1;

	return  (st1.st_rdev==st2.st_rdev) ? 0 : 1;
}

int mount_get_devinfo(char *devname, int *readwrite, char *mntbuf, int maxmntbuf, char *optbuf, int maxoptbuf, char *fsbuf, int maxfsbuf)
{
	char col_fs[FSA_MAX_FSNAMELEN];
	char delims[]=" \t\n";
	struct utsname suname;
	char col_dev[128];
	char col_mnt[128];
	char col_opt[128];
	char line[1024];
	char temp[2048];
	char *saveptr;
	char *result;
	int res;
	FILE *f;
	int i;

	// init
	res=uname(&suname);
	*readwrite=-1; // unknown
	memset(mntbuf, 0, sizeof(mntbuf));
	memset(optbuf, 0, sizeof(optbuf));
	
	if ((f=fopen("/proc/mounts","rb"))==NULL)
	{	sysprintf("Cannot open /proc/mounts\n");
		return 1;
	}

	while(!feof(f))
	{
		if (stream_readline(f, line, 1024)>1)
		{
			result=strtok_r(line, delims, &saveptr);
			col_dev[0]=col_mnt[0]=col_fs[0]=col_opt[0]=0;
			for(i=0; result != NULL && i<=3; i++)
			{
				switch (i) // only the second word is a mount-point
				{
					case 0:
						snprintf(col_dev, sizeof(col_dev), "%s", result);
						break;
					case 1:
						snprintf(col_mnt, sizeof(col_mnt), "%s", result);
						break;
					case 2:
						snprintf(col_fs, sizeof(col_fs), "%s", result);
						break;
					case 3:
						snprintf(col_opt, sizeof(col_opt), "%s", result);
						break;
				}
				result = strtok_r(NULL, delims, &saveptr);
			}
			
			if ((devcmp(col_dev, devname)==0) && (get_filesystem_info(col_dev, col_mnt, temp, sizeof(temp))==0))
			{
				*readwrite=get_filesystem_rw_status(col_opt);
				snprintf(mntbuf, maxmntbuf, "%s", col_mnt);
				snprintf(optbuf, maxoptbuf, "%s", col_opt);
				snprintf(fsbuf, maxfsbuf, "%s", col_fs);
				fclose(f);
				return 0;
			}
		}
	}

	fclose(f);
	return -1;
}

struct fsmount unixfs[] = 
{
	{"ext3", NULL},
	{"reiserfs", NULL},
	{"reiser4", NULL},
	{"xfs", "nouuid"},
	{"jfs", NULL},
	{"ext2", NULL},
	{"btrfs", NULL},
	{"ext4", NULL},
	{NULL, NULL},
};

int mount_partition(char *partition, char *mntbuf, char *fsbuf, int maxfsbuf, int flags)
{
	const char *mntopt;
	int i;
	
	// if fsbug is not empty, use the filesystem which is specified
	if (fsbuf[0])
	{
		mntopt=NULL;
		for (i=0; unixfs[i].filesys; i++)
			if (strcmp(fsbuf, unixfs[i].filesys)==0)
				mntopt=unixfs[i].mntopt;
		msgprintf(MSG_DEBUG1, "mount_partition(%s) = attempt with %s\n", partition, fsbuf);
		if (mount(partition, mntbuf, fsbuf, flags, mntopt)==0)
			return 0;
	}
	else // no filesystem specified
	{
		memset(fsbuf, 0, maxfsbuf);
		for (i=0; unixfs[i].filesys; i++)
		{
			msgprintf(MSG_DEBUG1, "mount_partition(%s) = attempt with %s with opt=[%s]\n", partition, unixfs[i].filesys, unixfs[i].mntopt);
			if (mount(partition, mntbuf, unixfs[i].filesys, flags, unixfs[i].mntopt)==0)
			{	snprintf(fsbuf, maxfsbuf, "%s", unixfs[i].filesys);
				return 0;
			}
		}
	}
	
	return -1;
}

int unmount_partition(char *mntbuf)
{
	msgprintf(MSG_DEBUG1, "unmount_partition(%s)\n", mntbuf);
	return umount2(mntbuf, 0);
}
