/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#ifndef __QUEUE_H__
#define __QUEUE_H__

#include <pthread.h>

enum	{QITEM_STATUS_NULL=0, QITEM_STATUS_TODO, QITEM_STATUS_PROGRESS, QITEM_STATUS_DONE};
enum	{QITEM_TYPE_NULL=0, QITEM_TYPE_BLOCK, QITEM_TYPE_DICO};

struct s_dico;

struct s_blockinfo
{
	char	*blkdata; // pointer to data block (compressed or uncompressed)
	u32	blkrealsize; // size of the data in the normal state (uncompressed)
	u64	blkoffset; // offset of the block in the normal file
	u32	blkarcsum; // checksum of the block as it it when it's in the archive (compressed)
	u64	blkarsize; // size of the block as it is in the archive (compressed)
	u32	blkcompalgo; // algo used to compressed the block
	bool	blklocked; // true if locked (being compressed)
	bool	blkcompressed; // true if the block has been processed by the compression thread
};

struct s_queueitem
{	// ------- common information: for all items of the queue --------------------------
	int 			type; // QITEM_TYPE_BLOCK or QITEM_TYPE_DICO
	s64			itemnum; // unique identifier of the item in the queue
	int			status; // compressed, being-compressed, not-yet-compressed
	struct s_queueitem	*next; // next block in the linked list
	// ------- information used when type==QITEM_TYPE_BLOCK (for blocks only) ----------
	struct s_blockinfo 	blkinfo; // extra info used for compressed blocks
	// ------- information used when type==QITEM_TYPE_DICO (for headers only) ----------
	char			magic[FSA_SIZEOF_MAGIC+1]; // magic for dico
	struct s_dico		*dico; // if (type==QITEM_TYPE_DICO)
};

struct s_queue
{
	struct s_queueitem	*head; // head of the queue: first item
	pthread_mutex_t		mutex; // pthread mutex for data protection
	pthread_cond_t		cond; // condition for pthread synchronization
	s64			curitemnum; // unique id given to every new item (block or header)
	u64			itemcount; // how many items there are (headers + blocks)
	u64			blkcount; // how many blocks items there are (items where type==QITEM_TYPE_BLOCK only)
	u64			blkmax; // how many blocks items there can be before the queue is considered as full
	bool			endofqueue; // set to true when no more data to put in queue (like eof): reader must stop
};

// ---- error management
enum
{	QERR_SUCCESS=0, // should not be used
	QERR_FAIL=-1, // any fatal error
	QERR_NOMEM=-2, // malloc failed, out of memory
	QERR_INVAL=-3, // a parameter is invalid (null pointer)
	QERR_NOTFOUND=-4, // item not found in the list
	QERR_ENDOFQUEUE=-5, // we have reached the end of the queue
	QERR_WRONGTYPE=-6, // the first item found in the queue is not what expected
	QERR_CLOSED=-7 // the queue has been closed by the reader: don't write anymore
};

char *qerr(s64 err);

// ----return status
// a) ">0" means success for functions that return item numbers
// b) =0   means success for functions that do not return item numbers
// c) "<0" QERR error number

// init and destroy
s64  queue_init(struct s_queue *l, s64 blkmax);
s64  queue_destroy(struct s_queue *l);

// information functions
s64  queue_count(struct s_queue *l);
s64  queue_count_status(struct s_queue *l, int status);
s64  queue_is_first_item_ready(struct s_queue *q);
s64  queue_check_next_item(struct s_queue *q, int *type, char *magic);

// modification functions
s64  queue_add_block(struct s_queue *q, struct s_blockinfo *blkinfo, int status);
s64  queue_add_header(struct s_queue *q, struct s_dico *d, char *magic);
s64  queue_replace_block(struct s_queue *q, s64 itemnum, struct s_blockinfo *blkinfo, int newstatus);
s64  queue_destroy_first_item(struct s_queue *q);

// end of queue functions
s64  queue_set_end_of_queue(struct s_queue *q, bool state);
bool queue_get_end_of_queue(struct s_queue *q);

// get item from queue functions
s64  queue_get_first_block_todo(struct s_queue *q, struct s_blockinfo *blkinfo);
s64  queue_dequeue_header(struct s_queue *q, struct s_dico **d, char *magicbuf);
s64  queue_dequeue_block(struct s_queue *q, struct s_blockinfo *blkinfo);
s64  queue_dequeue_first(struct s_queue *q, int *type, char *magic, struct s_dico **dico, struct s_blockinfo *blkinfo);

#endif // __QUEUE_H__
