/*
 * fsarchiver: Filesystem Archiver
 *
 * Copyright (C) 2008-2009 Francois Dupoux.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License v2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Homepage: http://www.fsarchiver.org
 */

#include "fsarchiver.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>

#include "archive.h"
#include "common.h"
#include "comp_gzip.h"
#include "comp_bzip2.h"
#include "comp_lzma.h"
#include "comp_lzo.h"
#include "syncthread.h"
#include "thread_comp.h"

int compress_block_generic(struct s_blockinfo *blkinfo)
{
	u64 compsize;
	char *bufdat;
	int res;
	u64 bufsize;
	
	bufsize = (blkinfo->blkrealsize) + (blkinfo->blkrealsize / 16) + 64 + 3; // alloc bigger buffer else lzo will crash
	if ((bufdat=malloc(bufsize))==NULL)
	{	errprintf("malloc(%ld) failed: out of memory\n", (long)blkinfo->blkrealsize);
		return -1;
	}
	
	// -------------- compress the block
	switch (g_options.compresslevel)
	{
#ifdef OPTION_LZO_SUPPORT
		case COMPRESS_LZO:
			res=compress_block_lzo(blkinfo->blkrealsize, &compsize, (u8*)blkinfo->blkdata, blkinfo->blkrealsize, (void*)bufdat, bufsize);
			blkinfo->blkcompalgo=COMPRESS_LZO;
			break;
#endif // OPTION_LZO_SUPPORT
		case COMPRESS_GZIP:
			res=compress_block_gzip(blkinfo->blkrealsize, &compsize, (u8*)blkinfo->blkdata, blkinfo->blkrealsize, (void*)bufdat, bufsize);
			blkinfo->blkcompalgo=COMPRESS_GZIP;
			break;
		case COMPRESS_BZIP2:
			res=compress_block_bzip2(blkinfo->blkrealsize, &compsize, (u8*)blkinfo->blkdata, blkinfo->blkrealsize, (void*)bufdat, bufsize);
			blkinfo->blkcompalgo=COMPRESS_BZIP2;
			break;
#ifdef OPTION_LZMA_SUPPORT
		case COMPRESS_LZMA:
			res=compress_block_lzma(blkinfo->blkrealsize, &compsize, (u8*)blkinfo->blkdata, blkinfo->blkrealsize, (void*)bufdat, bufsize);
			blkinfo->blkcompalgo=COMPRESS_LZMA;
			break;
#endif // OPTION_LZMA_SUPPORT
		default:
			free(bufdat);
			msgprintf(2, "invalid compression level: %d\n", (int)g_options.compresslevel);
			return -1;
	}
	
	// check compression status and efficiency
	if ((res==0) && (compsize < blkinfo->blkrealsize)) // compression worked and was good
	{	free(blkinfo->blkdata); // free old buffer (with uncompressed data)
		blkinfo->blkdata=bufdat; // new buffer (with compressed data)
		blkinfo->blkarsize=compsize;
		blkinfo->blkcompressed=true;
	}
	else // compressed version is bigger: keep the original block
	{	memcpy(bufdat, blkinfo->blkdata, blkinfo->blkrealsize);
		free(blkinfo->blkdata); // free old buffer
		blkinfo->blkdata=bufdat; // new buffer
		blkinfo->blkarsize=blkinfo->blkrealsize;
		blkinfo->blkcompalgo=COMPRESS_NONE;
		blkinfo->blkcompressed=false;
	}
	
	// calculates the block checksum
	blkinfo->blkarcsum=fletcher32((void*)blkinfo->blkdata, blkinfo->blkarsize);
	
	return 0;
}

int decompress_block_generic(struct s_blockinfo *blkinfo)
{
	u64 checkorigsize;
	char *bufdat;
	int res;
	
	// ---- allocate memory for uncompressed data
	if ((bufdat=malloc(blkinfo->blkrealsize))==NULL)
	{	errprintf("malloc(%ld) failed: cannot allocate memory for compressed block\n", (long)blkinfo->blkrealsize);
		return -1;
	}
	
	// check the block checksum
	if (fletcher32((u8*)blkinfo->blkdata, blkinfo->blkarsize)!=(blkinfo->blkarcsum))
	{	errprintf("block is corrupt at blockoffset=%ld, blksize=%ld\n", (long)blkinfo->blkoffset, (long)blkinfo->blkrealsize);
		memset(bufdat, 0, blkinfo->blkrealsize);
	}
	else // data not corrupted, decompresses the block
	{
		switch (blkinfo->blkcompalgo)
		{
			case COMPRESS_NONE:
				memcpy(bufdat, blkinfo->blkdata, blkinfo->blkarsize);
				res=0;
				break;
#ifdef OPTION_LZO_SUPPORT
			case COMPRESS_LZO:
				if ((res=uncompress_block_lzo(blkinfo->blkarsize, &checkorigsize, (void*)bufdat, blkinfo->blkrealsize, (u8*)blkinfo->blkdata))!=0)
				{	errprintf("uncompress_block_lzo()=%d failed: finalsize=%ld and checkorigsize=%ld\n", 
						res, (long)blkinfo->blkarsize, (long)checkorigsize);
					memset(bufdat, 0, blkinfo->blkrealsize);
					// TODO: inc(error_counter);
				}
				break;
#endif // OPTION_LZO_SUPPORT
			case COMPRESS_GZIP:
				if ((res=uncompress_block_gzip(blkinfo->blkarsize, &checkorigsize, (void*)bufdat, blkinfo->blkrealsize, (u8*)blkinfo->blkdata))!=0)
				{	errprintf("uncompress_block_gzip()=%d failed: finalsize=%ld and checkorigsize=%ld\n", 
						res, (long)blkinfo->blkarsize, (long)checkorigsize);
					memset(bufdat, 0, blkinfo->blkrealsize);
					// TODO: inc(error_counter);
				}
				break;
			case COMPRESS_BZIP2:
				if ((res=uncompress_block_bzip2(blkinfo->blkarsize, &checkorigsize, (void*)bufdat, blkinfo->blkrealsize, (u8*)blkinfo->blkdata))!=0)
				{	errprintf("uncompress_block_bzip2()=%d failed: finalsize=%ld and checkorigsize=%ld\n", 
						res, (long)blkinfo->blkarsize, (long)checkorigsize);
					memset(bufdat, 0, blkinfo->blkrealsize);
					// TODO: inc(error_counter);
				}
				break;
#ifdef OPTION_LZMA_SUPPORT
			case COMPRESS_LZMA:
				if ((res=uncompress_block_lzma(blkinfo->blkarsize, &checkorigsize, (void*)bufdat, blkinfo->blkrealsize, (u8*)blkinfo->blkdata))!=0)
				{	errprintf("uncompress_block_lzma()=%d failed: finalsize=%ld and checkorigsize=%ld\n", 
						res, (long)blkinfo->blkarsize, (long)checkorigsize);
					memset(bufdat, 0, blkinfo->blkrealsize);
					// TODO: inc(error_counter);
				}
				break;
#endif // OPTION_LZMA_SUPPORT
			default:
				errprintf("unsupported compression algorithm: %ld\n", (long)blkinfo->blkcompalgo);
				return -1;
		}
		if (res==0)
		{
			free(blkinfo->blkdata); // free old buffer (with compressed data)
			blkinfo->blkdata=bufdat; // pointer to new buffer with uncompressed data
		}
	}
	
	return 0;
}

int compression_function(struct s_archive *ai, int oper) //COMPTHR_COMPRESS, COMPTHR_DECOMPRESS
{
	struct s_blockinfo blkinfo;
	s64 blknum;
	int res;
	
	if (ai==NULL)
	{	errprintf("ai is NULL\n");
		goto thread_comp_fct_error;
	}
	
	while (queue_get_end_of_queue(&g_queue)==false)
	{
		if ((blknum=queue_get_first_block_todo(&g_queue, &blkinfo))>0) // block found
		{
			switch (oper)
			{
				case COMPTHR_COMPRESS:
					res=compress_block_generic(&blkinfo);
					break;
				case COMPTHR_DECOMPRESS:
					res=decompress_block_generic(&blkinfo);
					break;
				default:
					errprintf("oper is invalid: %d\n", oper);
					goto thread_comp_fct_error;
			}
			if (res!=0)
			{	errprintf("compress_block()=%d failed\n", res);
				goto thread_comp_fct_error;
			}
			// don't check for errors: it's normal to fail when we terminate after a problem
			queue_replace_block(&g_queue, blknum, &blkinfo, QITEM_STATUS_DONE);
		}
	}
	
	/*struct timespec t1, t2, t3, t4, t5, t6;
	u64 delaya, delayb, delayc;
	clock_gettime(CLOCK_REALTIME, &t1);
	clock_gettime(CLOCK_REALTIME, &t2);
	errprintf("comp-thread=%ld: =====================================\n", threadid);
	errprintf("comp-thread=%ld: TOTAL time in queue_get_first_block_todo()=%ld\n", threadid, tota);
	errprintf("comp-thread=%ld: TOTAL time in compress_block_generic()=%ld\n", threadid, totb);
	errprintf("comp-thread=%ld: TOTAL time in queue_replace_block()=%ld\n", threadid, totc);
	errprintf("comp-thread=%ld: =====================================\n", threadid);*/
	
	msgprintf(MSG_DEBUG1, "THREAD-COMP: exit success\n");
	return 0;
	
thread_comp_fct_error:
	get_stopfillqueue();
	msgprintf(MSG_DEBUG1, "THREAD-COMP: exit error\n");
	return 0;
}

void *thread_comp_fct(void *args)
{
	inc_secthreads();
	compression_function((struct s_archive *)args, COMPTHR_COMPRESS);
	dec_secthreads();
	return NULL;
}

void *thread_decomp_fct(void *args)
{
	inc_secthreads();
	compression_function((struct s_archive *)args, COMPTHR_DECOMPRESS);
	dec_secthreads();
	return NULL;
}
